/*
 * BSD Licence:
 * Copyright (c) 2001, Ben Houston [ ben@exocortex.org ]
 * Exocortex Technologies [ www.exocortex.org ]
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright 
 * notice, this list of conditions and the following disclaimer in the 
 * documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the <ORGANIZATION> nor the names of its contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */


using System;
using System.Diagnostics;
using System.Collections;
using System.Drawing;
using System.IO;
using System.Runtime.Serialization;
using System.Xml.Serialization;
using System.Windows.Forms;
using System.Xml;

using Exocortex;
using Exocortex.Collections;
using Exocortex.Collections.Specialized;
using Exocortex.Geometry3D;
using Exocortex.OpenGL;
using Exocortex.Cg;
using Exocortex.Windows.Forms;
			  
using ExoEngine.Geometry;
using ExoEngine.BSPTree;

namespace ExoEngine.Rendering {

	public class CgGouraudRenderer : Renderer {
	
		public CgGouraudRenderer() {
		}

		public override void	Initialize( World world ) {
			base.Initialize( world );

			if( CgUtils.GetCgCompilerPath() == null ) {
				ExoEngine.MessageBox( "Could not find the Cg Compiler, cgc.exe.\n" +
					"Please install the Visual Studio .NET Tutorials from NVIDIA's website:\n" +
					@"http://developer.nvidia.com/view.asp?IO=cg_tutorials", "Cg Not Installed / Not Available" );
				return;
			}

			Debug2.Push();

			//Trace2.WriteVar( "CGUtils.InitGLExtensions()", success );
			_cgContext = new CgContext();
			StreamReader streamReader = new StreamReader( ResourceLoader.GetStream( this.GetType(), "CgGouraudRenderer.cg" ) );
			string contents = streamReader.ReadToEnd();
			streamReader.Close();
			StreamWriter streamWriter = new StreamWriter( "CgGouraudRenderer.cg", false );
			streamWriter.WriteLine( contents );
			streamWriter.Close();

			string cgCompilerPath = CgUtils.GetCgCompilerPath();
			Debug2.WriteLine( "cgCompilerPath = " + ( ( cgCompilerPath != null ) ? cgCompilerPath : "(null)" ) );

			_cgContext.AddProgramFromFile( "CgGouraudRenderer.cg", CgProfileType.Vertex );
			_programMain = _cgContext.GetProgramByName( "main" );

			// Generate an ID for the program
			// Feed the compiled program to the graphics hardware
			_programMain.Load( 1 );

			// Now for any program parameter, we associate (or bind) a pointer (or iterator)
			// that will from now on be used to refer to this parameter

			// Bind the program uniform parameters (same for every vertex)
			_bindModelViewProj	= _programMain.GetBindByName( "ModelViewProj" );    
			_bindLightPos		= _programMain.GetBindByName( "LightPos" );
			_bindEyePos			= _programMain.GetBindByName( "EyePos" );    
			_bindMaterialColor	= _programMain.GetBindByName( "MaterialColor" );
			_bindLightColor		= _programMain.GetBindByName( "LightColor" );

			_bindKa			= _programMain.GetBindByName( "Ka" );
			_bindKd			= _programMain.GetBindByName( "Kd" );
			_bindKs			= _programMain.GetBindByName( "Ks" );
			_bindShininess	= _programMain.GetBindByName( "Shininess" );

			// Bind the program varying parameters (varying per vertex)
			_bindINPosition		= _programMain.GetBindByName( "IN.Position" );
			_bindINNormal		= _programMain.GetBindByName( "IN.Normal" );
			_bindINTextureCoord	= _programMain.GetBindByName( "IN.TexCoord" );

			// Enable client states
			Debug2.WriteLine( "enabling 'position' client state..." );
			_bindINPosition.EnableClientState();

			Debug2.WriteLine( "enabling 'normal' client state..." );
			_bindINNormal.EnableClientState();			

			Debug2.WriteLine( "enabling 'texture coord' client state..." );
			_bindINTextureCoord.EnableClientState();			

			Debug2.Pop();
		}

		public override void	Shutdown( World world ) {
			base.Shutdown( world );
			if( _cgContext == null ) {
				return;
			}
			Debug2.Push();

			// Disable client states
			Debug2.WriteLine( "disabling 'texture coord' client state..." );
			_bindINTextureCoord.DisableClientState();

			Debug2.WriteLine( "disabling 'position' client state..." );
			_bindINNormal.DisableClientState();

			Debug2.WriteLine( "disabling 'normal' client state..." );
			_bindINPosition.DisableClientState();

			_bindINNormal.Dispose();
			_bindINPosition.Dispose();

			_bindShininess.Dispose();
			_bindKs.Dispose();
			_bindKd.Dispose();

			_bindLightColor.Dispose();
			_bindMaterialColor.Dispose();
			_bindEyePos.Dispose();
			_bindLightPos.Dispose();
			_bindModelViewProj.Dispose();


			// Bind the program varying parameters (varying per vertex)

			// Free program iterator
			Debug2.WriteLine( "freeing program..." );
			_programMain.Dispose();

			// Free context
			Debug2.WriteLine( "freeing context..." );
			_cgContext.Dispose();

			// Final cleanup
			Debug2.WriteLine( "cleaning up..." );
			CgUtils.Cleanup();

			Debug2.Pop();
		}

		//---------------------------------------------------------------------------------------------------------
		//---------------------------------------------------------------------------------------------------------
		
		protected	CgContext	_cgContext = null;
		protected	CgProgram	_programMain		= null;

		protected	CgBind		_bindModelViewProj	= null;
		protected	CgBind		_bindLightPos		= null;
		protected	CgBind		_bindEyePos			= null;
		protected	CgBind		_bindMaterialColor	= null;
		protected	CgBind		_bindLightColor		= null;
		protected	CgBind		_bindKa				= null;
		protected	CgBind		_bindKd				= null;
		protected	CgBind		_bindKs				= null;
		protected	CgBind		_bindShininess		= null;
		protected	CgBind		_bindINPosition		= null;
		protected	CgBind		_bindINNormal		= null;
		protected	CgBind		_bindINTextureCoord	= null;
		
		//---------------------------------------------------------------------------------------------------------
		//---------------------------------------------------------------------------------------------------------

		
		public override unsafe void Render( World world, Faces faces, RenderSettings settings ) {
			base.Render( world, faces, settings );
			if( _cgContext == null ) {
				return;
			}

			_programMain.Bind();
			_cgContext.EnableProgramType( CgProfileType.Vertex );

			// Transform from model space to clip space
			_bindModelViewProj.BindUniformStateMatrix( CgMatrixType.ModelViewProjection, CgMatrixFormat.Identity );

			// Ambient coefficient
			_bindKs.BindUniform4f( 0.2f );
			// Diffuse coefficient
			_bindKd.BindUniform4f( 1f );
			// Specular coefficient
			_bindKs.BindUniform4f( 1f );
			// Specular exponent
			_bindShininess.BindUniform4f( 30 );

			Vector3 cameraLocation = world.Camera.Translation;
			Vector3 lightLocation = world.Light;
			_bindEyePos.BindUniform4f( cameraLocation.X, cameraLocation.Y, cameraLocation.Z );
			_bindLightPos.BindUniform4f( lightLocation.X, lightLocation.Y, lightLocation.Z );
			_bindLightColor.BindUniform4f( Color.White );

			GL.glDisable( GLOption.Blend );
			GL.glEnable( GLOption.ColorMaterial );
			if( settings.Textures ) {
				GL.glEnable( GLOption.Texture2D );
			}
			else {
				GL.glDisable( GLOption.Texture2D );
			}

			if( settings.ZBuffer ) {
				GL.glEnable( GLOption.DepthTest );
			}
			else {
				GL.glDisable( GLOption.DepthTest );
			}
			if( settings.Wireframe ) {
				GL.glPolygonMode( GLPolygonFaces.Front, GLPolygonMode.Line );
			}
			else {
				GL.glPolygonMode( GLPolygonFaces.Front, GLPolygonMode.Fill );
			}

			GL.glCullFace( GLPolygonFaces.Back );
			
			foreach( Face face in faces ) {
				Vector3Collection	normalsArray	= null;
				if( face.IsVertexNormals == true ) {
					normalsArray = face.VertexNormals;
				}
				else {
					normalsArray = new Vector3Collection();
					Vector3		normal			= face.Normal;
					for( int i = 0; i < face.Points.Count; i ++ ) {
						normalsArray.Add( normal );
					}
				}

			
				UInt32Collection	facesArray			= new UInt32Collection();
				for( int i = 1; i < face.Points.Count - 1; i ++ ) {
					facesArray.Add( (uint) 0 );
					facesArray.Add( (uint) ( 0 + i ) );
					facesArray.Add( (uint) ( 0 + i + 1 ) );
				}

				if( settings.Textures ) {
					if ( face.IsTexture ) {
						GL.glEnable( GL.GL_TEXTURE_2D );
						GL.glBindTexture( GLTexture.Texture2D, face.Texture.OpenGLHandle );
					}
					else {
						GL.glDisable( GL.GL_TEXTURE_2D );
					}
				}
			
				int			vertexCount = face.Points.Count;
				int			faceCount	= ( facesArray.Count / 3 );

				Vector3[]	vertices		= face.Points.Data;
				Vector3[]	normals			= normalsArray.Data;
				Vector3[]	textureCoords	= face.TextureCoords.Data;
				uint[]		faceIndices		= facesArray.Data;

				_bindMaterialColor.BindUniform4f( settings.FaceColors ? face.Color : Color.White );

				fixed( Vector3* pVertices = & (vertices[0]),
						   pNormals = & (normals[0]),
						   pTextureCoords = & (textureCoords[0]) ) {

					this._bindINPosition.BindVaryingPointer( 3, pVertices );
					this._bindINNormal.BindVaryingPointer( 3, pNormals );
					this._bindINTextureCoord.BindVaryingPointer( 3, pTextureCoords );
					
					fixed( uint* pFaceIndices = & (faceIndices[0]) ) {
					
						GL.glDrawElements( GLPrimative.Triangles, faceCount * 3, GLUnsignedDataType.UnsignedInt, pFaceIndices );
					}

					GL.glErrorCheck();
				}
			}

			_cgContext.DisableProgramType( CgProfileType.Vertex );

		}
	}

}
