using System;
using System.Collections;
using System.Diagnostics;


namespace Exocortex.Collections.Specialized {
	/***begin code***/

	using T				= System.UInt32;
	using TCollection	= UInt32Collection;
	using TEnumerator	= UInt32Enumerator;

	//=====================================================================================
	//=====================================================================================

	
	/// <summary>
	/// A collection of planes
	/// </summary>
	[Serializable]
	public class UInt32Collection : IList, ICollection, IEnumerable, ICloneable {

		
		//---------------------------------------------------------------------------------
		// diagnostic methods

		static protected int s_maxTotalSize = 0;
		static protected int s_totalSize = 0;
		protected void	TrackMemoryUsage( int change ) {
			s_totalSize += change;
			int oldMaxTotalSize = s_maxTotalSize;
			s_maxTotalSize = Math.Max( s_totalSize, s_maxTotalSize );
			if( oldMaxTotalSize != s_maxTotalSize ) {
			//	Debug.WriteLine( "UInt64Collection s_maxTotalSize: " + ((( s_maxTotalSize * 8 )/10000)/100f) + " MB" );
			}
			//Debug.WriteLine( change + "\t(" + s_totalSize + ")\t" + Debug2.GetCallingMethod( 1 ) );
		}

		static public void OutputMemoryStats() {
			Debug.WriteLine( "MaxMemoryUsage: " + s_maxTotalSize );
			Debug.WriteLine( "MaxUInt32Collections: " + s_maxCollections );
		}

		public T[]	AllocArray( int newSize ) {
			TrackMemoryUsage( newSize );
			return new T[ newSize ];
		}
		public T[]	AllocArray( T[] newData ) {
			TrackMemoryUsage( newData.Length );
			return newData;
		}
		public void	ReplaceDataArray( T[] newData ) {
			Debug.Assert( newData != null );
			TrackMemoryUsage( - data.Length );
			this.data = newData;
		}
		public void	ReleaseArray( T[] oldData ) {
			Debug.Assert( oldData != null );
			TrackMemoryUsage( - oldData.Length );
		}

		//---------------------------------------------------------------------------------
		// members

		const	int		sDefaultCapacity	= 16;
		const	float	sGrowthFactor		= 1.3f;

		private	T[]		data;
		private	int		count;

		//---------------------------------------------------------------------------------
		// constructors

		static public int s_maxCollections = 0;
		static public int s_collections = 0;

		public void	NewCollection() {
			s_collections ++;
			int oldMax = s_maxCollections;
			s_maxCollections = Math.Max( s_collections, s_maxCollections );
			//Debug.WriteLine( "+ " + s_collections + "\t" + Debug2.GetCallingMethod( 0 ) );
			if( oldMax != s_maxCollections ) {
			//	Debug.WriteLine( "UInt64Collection s_maxCollections: " + s_maxCollections );
			}
		}
		public void	ReleaseCollection() {
			s_collections --;
			//Debug.WriteLine( "- " + s_collections + "\t" + Debug2.GetCallingMethod( 0 ) );
		}
		
		protected UInt32Collection( int initialCount, T[] initialData ) {
			Debug.Assert( initialCount >= 0 );
			Debug.Assert( initialData != null );
			Debug.Assert( initialData.Length >= initialCount );
			AllocArray( initialData );
			this.data	= initialData;
			this.count	= initialCount;

			NewCollection();
		}

		public	UInt32Collection() {
			this.data = AllocArray( sDefaultCapacity );

			NewCollection();
		}

		public	UInt32Collection( int initialCapacity ) {
			Debug.Assert( initialCapacity >= 0 );
			this.data = AllocArray(  Math.Max( initialCapacity, sDefaultCapacity ) );
		
			NewCollection();
		}

		public	UInt32Collection( TCollection collection ) {
			Debug.Assert( collection != null );
			this.data = AllocArray( Math.Max( (int) Math.Ceiling( collection.Count * sGrowthFactor ), sDefaultCapacity ) );
			this.AddRange( collection );

			NewCollection();
		}

		public	UInt32Collection( T[] array ) {
			Debug.Assert( array != null );
			this.data = AllocArray( Math.Max( (int) Math.Ceiling( array.Length * sGrowthFactor ), sDefaultCapacity ) );
			this.AddRange( array );

			NewCollection();
		}

		~UInt32Collection() {
			ReleaseArray( this.data );
			ReleaseCollection();
		}

		//---------------------------------------------------------------------------------
		// properties

		public	T	this[ int index ] {
			get {
				Debug.Assert( 0 <= index && index < count );
				return this.data[ index ];
			}
			set {
				Debug.Assert( 0 <= index && index < count );
				this.data[ index ] = value;
			}
		}

		public int Count {
			get	{	
				return this.count;	
			}
		}

		public	T[]	Data {
			get {
				return	this.data;
			}
		}

		//---------------------------------------------------------------------------------
		// methods

		public void SetToArray( TCollection c ) {
			this.data = null;
			this.count = 0;

			// aggressive memory management
			//System.GC.Collect();

			this.data = new T[ c.Count ];
			this.count = c.Count;
			Array.Copy( c.data, 0, this.data, 0, c.Count );
		}
		
		public	TCollection	Clone() {
			return	new TCollection( this.count, (T[]) this.data.Clone() );
		}
		
		public  int	CurrentCapacity {
			get {
				if( this.data != null ) {
					return	this.data.Length;
				}
				return	0;
			}
		}

		private void EnsureCapacity( int requiredCapacity ) {
			Debug.Assert( requiredCapacity >= 0 );

			int currentCapacity = this.CurrentCapacity;

			// if required capacity is already met -> do nothing
			if( requiredCapacity <= currentCapacity ) {
				return;
			}

			SetCapacity( Math.Max( (int) Math.Ceiling( requiredCapacity * sGrowthFactor ), sDefaultCapacity ) );
		}

		public void SetCapacity( int requiredCapacity ) {
			Debug.Assert( requiredCapacity >= 0 );

			int currentCapacity = this.CurrentCapacity;
		
			// if capacity already equals required -> do nothing
			if( currentCapacity == requiredCapacity ) {
				return;
			}

			Debug.Assert( currentCapacity < requiredCapacity );

			T[] newData = AllocArray( requiredCapacity );
			if( this.data != null ) {
				Array.Copy( this.data, 0, newData, 0, currentCapacity );
			}
			ReplaceDataArray( newData );
	
			// aggressive memory management
			//System.GC.Collect();
		}

		public	void	CopyCollection( TCollection col ) {
			this.Clear();
			//this.count = 0;
			//this.TrimToSize();

			//System.GC.Collect();

			this.EnsureCapacity( col.Count );
			this.count = col.count;
			Array.Copy( col.data, 0, this.data, 0, this.count );
		}
		
		public void TrimToSize() {
			if( this.count == 0 ) {
				this.data = new T[0];
				return;
			}
			if( this.count == this.data.Length ) {
				return;
			}

			int oldSize = this.data.Length;
			T[] newData = AllocArray( this.count );
			Array.Copy( this.data, 0, newData, 0, this.count );
			ReplaceDataArray( newData );
	
			// aggressive memory management
			//System.GC.Collect();
		}

		public void Clear(){
			this.count	= 0;
			//this.data	= null;
		}
		
		public int Add( T element ) {
			EnsureCapacity( this.count + 1 );
			this.data[ this.count ] = element;
			return this.count ++;    // BUG: should size be the place it was inserted or the length of the array?
		}

		public void AddRange( T[] array ) {
			Debug.Assert( array != null );
			if( array.Length > 0 ) {
				EnsureCapacity( this.count + array.Length );
				Array.Copy( array, 0, this.data, this.count, array.Length );
				this.count += array.Length;
			}
		}

		public void AddRange( TCollection collection ) {
			Debug.Assert( collection != null );
			if( collection.Count > 0 ) {
				EnsureCapacity( this.count + collection.count );
				Array.Copy( collection.data, 0, this.data, this.count, collection.count );
				this.count += collection.count;
			}
		}

		public	T	RemoveLast() {
			Debug.Assert( this.Count > 0 );
			
			T lastElement = this[ this.Count - 1 ];
			this.count --;
			return	lastElement;
		}
		
		public bool Contains( T element ) {
			for( int i = 0; i < this.count; i++ ) {
				if( this.data[i] == element ) {
					return true;
				}
			}
			return false;
		}

		public void CopyTo( T[] array, int index ) {
			Array.Copy( this.data, 0, array, 0, this.count );
		}


		public int IndexOf( T element ) {
			for( int i = 0; i < this.count; i ++ ) {
				if( this.data[i] == element ) {
					return i;
				}
			}
			return -1;
		}

	
		public void Insert( int index, T element ) {
			// BUG: element may not fit into array????
			Debug.Assert( 0 <= index && index < this.count );

			if( this.count == this.data.Length) {
				EnsureCapacity( this.count + 1 );
			}
			if( index < this.count ) {
				Array.Copy( this.data, index, this.data, index + 1, this.count - index );
			}
			this.data[ index ] = element;
			this.count++;
	
			// aggressive memory management
			//System.GC.Collect();
		}
		
		public void Remove( T element ) {
			for( int i = 0; i < this.count; i++ ) {
				if( this.data[i] == element ) {
					this.RemoveAt( i );
					return;
				}
			}
			throw new ArgumentException( "cannot find element: " + element.ToString(), "element" );
		}

		public T	RemoveAt( int index ){
			Debug.Assert( 0 <= index && index < this.count );
			T element = data[ index ];
			T[] tail = AllocArray(  this.count - index );
			Array.Copy( this.data, index + 1, tail, 0, this.count - index - 1 );
			Array.Copy( tail, 0, this.data, index, this.count - index - 1 );
			this.count --;

			ReleaseArray( tail );
			
			// aggressive memory management
			//System.GC.Collect();

			return element;
		}

		public TEnumerator GetEnumerator() {
			return new TEnumerator( this );
		}
		
		public void Reverse() {
			for( int i = 0; i < this.count / 2; i ++ ) {
				T temp = this[ i ];
				this[i] = this[ this.count - i - 1 ];
				this[ this.count - i - 1 ] = temp;
			}
		}

		/*
		 * Old binary search function - May 23, 2002
		 * 
		private int	PrivateBinarySearch( int start, int end, T element ) {
			Debug.Assert( this.data != null );
			Debug.Assert( start < end );
			
			int middle = ( start + end ) >> 1;

			if( this.data[ middle ] == element ) {
				return	middle;
			}
			else if( start == middle ) {
				return	-1;
			}
			else if( this.data[ middle ] < element ) {
				return	PrivateBinarySearch( middle, end, element );
			}
			else {//if( data[ middle ] > element ) {
				return	PrivateBinarySearch( start, middle, element );
			}
		}
		*/

		public int PrivateBinarySearch( int index, int length, T value ) {
			Debug.Assert( this.data != null );
			Debug.Assert( index < length );

			int lo = index;
			int hi = index + length - 1;
			while ( lo <= hi ) {
				int i = ( lo + hi ) >> 1;
				T element = this.data[ i ];
				if ( element == value ) {
					return i;
				}
				if( element < value ) {
					lo = i + 1;
				}
				else {
					hi = i - 1;
				}
			}
			return -1;
		}

		public int BinarySearch( T element ) {
			if( this.data == null || this.count == 0 ) {
				return	-1;
			}

			return	this.PrivateBinarySearch( 0, this.count, element );
		}

		public int BinarySearch( T element, int index, int length ) {
			Debug.Assert( 0 <= index && index < this.count );
			Debug.Assert( length >= 0 );
			Debug.Assert( ( index + length ) <= this.count );

			if( this.data == null || this.count == 0 ) {
				return	-1;
			}

			return	this.PrivateBinarySearch( index, length, element );
		}

		public void	FromArray( T[] array ) {
			if( array != null ) {
				this.Clear();
				foreach( T o in array ) {
					this.Add( o );
				}
			}
		}
		
		public T[] ToArray() {
			T[] array = AllocArray( this.Count );
			this.CopyTo( array, 0 );
			return array;
		}

		public int RemoveDuplicates( IComparer comparer ) {
			if( this.count <= 1 ) {
				return 0;
			}
			int newIndex = 1, oldIndex;
			T lastUniqueElement = this.data[ 0 ];
			for( oldIndex = 1; oldIndex < this.count; oldIndex ++ ) {
				T currentElement = this.data[ oldIndex ];
				int compareResult = comparer.Compare( currentElement, lastUniqueElement );
				Debug.Assert( compareResult >= 0 );
				if( compareResult > 0 ) {
					this.data[ newIndex ++ ] = currentElement;
					lastUniqueElement = currentElement;
				}
			}
			this.count = newIndex;
			return	( oldIndex - newIndex );
		}

		public int RemoveDuplicates() {
			if( this.count <= 1 ) {
				return 0;
			}
			int newIndex = 1, oldIndex;
			T lastUniqueElement = this.data[ 0 ];
			for( oldIndex = 1; oldIndex < this.count; oldIndex ++ ) {
				T currentElement = this.data[ oldIndex ];
				if( currentElement != lastUniqueElement ) {
					this.data[ newIndex ++ ] = currentElement;
					lastUniqueElement = currentElement;
				}
			}
			this.count = newIndex;
			return	( oldIndex - newIndex );
		}

		/*
		 * Old sort function - May 23, 2002
		 * 
		public void Sort() {
			if( this.data != null ) {
				Array.Sort( this.data, 0, this.count );
			}
		}
		*/

		public void	Sort( IComparer comparer ) {
			System.Array.Sort( data, 0, count, comparer );
		}

		public void Sort() {
			if( this.data == null || this.count <= 1 ) {
				return;
			}
			this.PrivateQuickSort( 0, this.count - 1 );
		}

		private void PrivateQuickSort( int left, int right ) {
			// Can use the much faster jit helpers for array access.
			do {
				int i = left;
				int j = right;
				T x = this.data[ (i + j) >> 1 ];
				do {
					while( this.data[i] < x ) i++;
					while( x < this.data[j] ) j--;
					
					if( i > j ) {
						break;
					}
					if( i < j ) {
						// swap keys[i] and keys[j]
						T key = this.data[i];
						this.data[i] = this.data[j];
						this.data[j] = key;
					}
					i++;
					j--;
				} while( i <= j );

				if( j - left <= right - i ) {
					if( left < j ) {
						this.PrivateQuickSort( left, j );
					}
					left = i;
				}
				else {
					if( i < right ) {
						this.PrivateQuickSort( i, right );
					}
					right = j;
				}
			} while( left < right );
		}
		
		//---------------------------------------------------------------------------------
		// special constructors

		static public TCollection	MergeSortedCollections( TCollection a, TCollection b ) {
			Debug.Assert( a != null );
			Debug.Assert( b != null );

			// handle simple cases
			if( a.Count == 0 ) {
				return	b.Clone();
			}
			if( b.Count == 0 ) {
				return	a.Clone();
			}

			int aIndex = 0;
			int	bIndex = 0;
			int destIndex = 0;
			TCollection mergedCollection = new TCollection( a.Count + b.Count );
			mergedCollection.count = a.Count + b.Count;

			while ( ( aIndex < a.count ) && ( bIndex < b.count ) ) {
				if( a.data[ aIndex ] < b.data[ bIndex ] ) {
					mergedCollection.data[ destIndex ++ ] = a.data[ aIndex ++ ];
				}
				else {
					mergedCollection.data[ destIndex ++ ] = b.data[ bIndex ++ ];
				}
			}
				
			if( aIndex < a.Count ) {
				Debug.Assert( bIndex == b.Count ); 
				for( int i = aIndex; i < a.count; i ++ ) {
					mergedCollection.data[ destIndex ++ ] = a.data[ i ];
				}
			}
			if( bIndex < b.Count ) {
				Debug.Assert( aIndex == a.count ); 
				for( int i = bIndex; i < b.Count; i ++ ) {
					mergedCollection.data[ destIndex ++ ] = b.data[ i ];
				}
			}

			return	mergedCollection;
		}

		//---------------------------------------------------------------------------------
		// IList interface

		object IList.this[int index] {
			get {
				Debug.Assert( 0 <= index && index < this.count );
				return this.data[ index ];
			}
			set {
				Debug.Assert( 0 <= index && index < this.count );
				this.data[ index ] = (T) value;
			}
		}

		bool	IList.IsFixedSize	{	get	{	return false;	}	}
		bool	IList.IsReadOnly	{	get {	return false;	}	}

		int		IList.Add( object element )			{	return this.Add( (T)element );		}
		bool	IList.Contains( object element )	{	return this.Contains( (T)element );	}
		int		IList.IndexOf( object element )		{	return this.IndexOf( (T)element );	}

		void	IList.Insert( int index, object element )	{	this.Insert( index, (T)element );	}
		void	IList.Remove( object element )				{	this.Remove( (T)element );					}
		void	IList.RemoveAt( int index )					{	this.RemoveAt( index );				}

		//---------------------------------------------------------------------------------
		// ICollection interface


		bool	ICollection.IsSynchronized	{	get {	return this.data.IsSynchronized;	}	}
		object	ICollection.SyncRoot		{	get {	return this.data.SyncRoot;			}	}

		void	ICollection.CopyTo( Array array, int index ) {
			Array.Copy( this.data, 0, array, index, this.count );
		}

		//---------------------------------------------------------------------------------
		// IEnumerable interface

		IEnumerator	IEnumerable.GetEnumerator()	{	return new TEnumerator( this );	}

		//---------------------------------------------------------------------------------
		// ICloneable interface

		object	ICloneable.Clone(){
			return	new TCollection( this.count, (T[]) this.data.Clone() );
		}
		
		//---------------------------------------------------------------------------------
	}

	//=====================================================================================
	//=====================================================================================

	public class UInt32Enumerator : IEnumerator {

		private TCollection		collection;
		private T				currentElement;
		private int				index;

		public UInt32Enumerator( TCollection collection ) {
			this.collection	= collection;
			this.index		= -1;
		}

		public void Reset() {
			this.index			= -1;
			this.currentElement	= (T) 0;
		}

		public T Current {
			get {
				return this.currentElement;
			}
		}

		object IEnumerator.Current {
			get {
				return this.currentElement;
			}
		}

		public bool MoveNext() {
			index ++;
			if( index < collection.Count ) {
				currentElement = collection[ index ];
				return true;
			}
			else {
				return false;
			}
		}
	}

	//=====================================================================================
	//=====================================================================================

}
