using System;
using System.Collections;
using System.Diagnostics;


namespace Exocortex.Collections.Specialized {
	/***begin code***/

	using T				= System.UInt64;
	using TCollection	= UInt40Collection;

	//=====================================================================================
	//=====================================================================================

	/// <summary>
	/// A collection of UInt40's
	/// </summary>
	[Serializable]
	public class UInt40Collection {

		//---------------------------------------------------------------------------------
		// diagnostic methods

		static protected int s_maxTotalSize = 0;
		static protected int s_totalSize = 0;
		protected void	TrackMemoryUsage( int change ) {
			s_totalSize += change;
			int oldMaxTotalSize = s_maxTotalSize;
			s_maxTotalSize = Math.Max( s_totalSize, s_maxTotalSize );
			if( oldMaxTotalSize != s_maxTotalSize ) {
				Debug.WriteLine( "UInt40Collection s_maxTotalSize: " + ((( s_maxTotalSize * 8 )/10000)/100f) + " MB" );
			}
			//Debug.WriteLine( change + "\t(" + s_totalSize + ")\t" + Debug2.GetCallingMethod( 1 ) );
		}

		static public void OutputMemoryStats() {
			Debug.WriteLine( "MaxMemoryUsage: " + s_maxTotalSize );
			Debug.WriteLine( "MaxUInt40Collections: " + s_maxCollections );
		}

		public T[]	AllocArray( int newSize ) {
			TrackMemoryUsage( newSize );
			return UInt40.CreateArray( newSize );
		}
		public T[]	AllocArray( T[] newData ) {
			TrackMemoryUsage( UInt40.GetSize40( newData ) );
			return newData;
		}
		public void	ReplaceDataArray( T[] newData ) {
			Debug.Assert( newData != null );
			TrackMemoryUsage( - UInt40.GetSize40( this.data64 ) );
			this.data64 = newData;
		}
		public void	ReleaseArray( T[] oldData ) {
			Debug.Assert( oldData != null );
			TrackMemoryUsage( - UInt40.GetSize40( oldData ) );
		}

		//---------------------------------------------------------------------------------
		// members

		const	int		sDefaultCapacity	= 16;
		const	float	sGrowthFactor		= 1.3f;

		private	T[]		data64;
		private	int		count40;

		//---------------------------------------------------------------------------------
		// constructors

		static private int s_maxCollections = 0;
		static private int s_collections = 0;

		static private void	NewCollection() {
			s_collections ++;
			int oldMax = s_maxCollections;
			s_maxCollections = Math.Max( s_collections, s_maxCollections );
			//Debug.WriteLine( "+ " + s_collections + "\t" + Debug2.GetCallingMethod( 0 ) );
			if( oldMax != s_maxCollections ) {
			//	Debug.WriteLine( "UInt64Collection s_maxCollections: " + s_maxCollections );
			}
		}
		static private void	ReleaseCollection() {
			s_collections --;
			//Debug.WriteLine( "- " + s_collections + "\t" + Debug2.GetCallingMethod( 0 ) );
		}

		//---------------------------------------------------------------------------------

		public	UInt40Collection() {
			this.data64 = AllocArray( sDefaultCapacity );

			NewCollection();
		}

		public	UInt40Collection( int initialCapacity ) {
			Debug.Assert( initialCapacity >= 0 );
			this.data64 = AllocArray( Math.Max( initialCapacity, sDefaultCapacity ) );
		
			NewCollection();
		}

		~UInt40Collection() {
			ReleaseArray( this.data64 );
			ReleaseCollection();
		}

		//---------------------------------------------------------------------------------

		public	T	this[ int index ] {
			get {
				Debug.Assert( 0 <= index && index < this.Count );
				return UInt40.GetElement( this.data64, index ); // this.data[ index ];
			}
			set {
				Debug.Assert( 0 <= index && index < this.Count );
				UInt40.SetElement( this.data64, index, value ); //this.data[ index ] = value;
			}
		}

		public int Count {
			get	{	
				return this.count40;	
			}
		}

		public int	CurrentCapacity {
			get {
				if( this.data64 != null ) {
					return	UInt40.GetSize40( this.data64 );
				}
				return	0;
			}
		}
		
		//---------------------------------------------------------------------------------
		
		private void EnsureCapacity( int requiredCapacity ) {
			Debug.Assert( requiredCapacity >= 0 );

			int currentCapacity = this.CurrentCapacity;

			// if required capacity is already met -> do nothing
			if( requiredCapacity <= currentCapacity ) {
				return;
			}

			SetCapacity( Math.Max( (int) Math.Ceiling( requiredCapacity * sGrowthFactor ), sDefaultCapacity ) );
			//System.GC.Collect();
		}

		public void SetCapacity( int requiredCapacity ) {
			Debug.Assert( requiredCapacity >= 0 );

			int currentCapacity = this.CurrentCapacity;
		
			// if capacity already equals required -> do nothing
			if( currentCapacity == requiredCapacity ) {
				return;
			}

			Debug.Assert( currentCapacity < requiredCapacity );

			T[] newData = AllocArray( requiredCapacity );
			if( this.data64 != null ) {
				Array.Copy( this.data64, 0, newData, 0, UInt40.GetSize64FromCount40( this.count40 ) );
			}
			ReplaceDataArray( newData );
	
			// aggressive memory management
			//System.GC.Collect();
		}

		public	void	CopyCollection( TCollection col ) {
			this.Clear();
			//this.count = 0;
			//this.TrimToSize();

			//System.GC.Collect();

			this.EnsureCapacity( col.Count );
			this.count40 = col.count40;
			Array.Copy( col.data64, 0, this.data64, 0, UInt40.GetSize64FromCount40( this.count40 ) );
		}

		//---------------------------------------------------------------------------------
	
		public int Add( T element ) {
			EnsureCapacity( this.Count + 1 );
			this.count40 ++;    
			this[ this.Count - 1 ] = element;
			return  this.Count;	// BUG: should return value be the place it was inserted or the length of the array?
		}

		public void AddRange( T[] array ) {
			Debug.Assert( array != null );
			EnsureCapacity( this.Count + array.Length );
			for( int i = 0; i < array.Length; i ++ ) {
				this.Add( array[ i ] );
			}
		}

		public void AddRange( TCollection collection ) {
			Debug.Assert( collection != null );
			EnsureCapacity( this.Count + collection.Count );
			for( int i = 0; i < collection.Count; i ++ ) {
				this.Add( collection[ i ] );
			}
		}

		public	T	RemoveLast() {
			Debug.Assert( this.Count > 0 );
			T lastElement = this[ this.Count - 1 ];
			this.count40 --;
			return	lastElement;
		}

		//---------------------------------------------------------------------------------
				
		public void TrimToSize() {
			if( this.Count == this.CurrentCapacity ) {
				return;
			}

			//int oldSize = UInt40.GetSize40( this.data.Length;
			T[] newData = AllocArray( this.Count );
			Array.Copy( this.data64, 0, newData, 0, UInt40.GetSize64FromCount40( this.count40 ) );
			ReplaceDataArray( newData );
	
			// aggressive memory management
			//System.GC.Collect();
		}

		public void Clear(){
			this.count40	= 0;
			//this.data	= null;
		}

		//---------------------------------------------------------------------------------

		public bool Contains( T element ) {
			for( int i = 0; i < this.Count; i++ ) {
				if( this[ i ] == element ) {
					return true;
				}
			}
			return false;
		}

		public int IndexOf( T element ) {
			for( int i = 0; i < this.Count; i ++ ) {
				if( this[ i ] == element ) {
					return i;
				}
			}
			return -1;
		}
	
		//---------------------------------------------------------------------------------

		public int PrivateBinarySearch( int index, int length, T value ) {
			Debug.Assert( this.data64 != null );
			//Debug.Assert( length > 0 );

			int lo = index;
			int hi = index + length - 1;
			while ( lo <= hi ) {
				int i = ( lo + hi ) >> 1;
				T element = this[ i ];
				if ( element == value ) {
					return i;
				}
				if( element < value ) {
					lo = i + 1;
				}
				else {
					hi = i - 1;
				}
			}
			return -1;
		}

		public int BinarySearch( T element ) {
			if( this.data64 == null || this.Count == 0 ) {
				return	-1;
			}

			return	this.PrivateBinarySearch( 0, this.Count, element );
		}

		public int BinarySearch( T element, int index, int length ) {
			Debug.Assert( ( 0 <= index && index < this.Count ) || ( index == 0 && length == 0 ) );
			Debug.Assert( length >= 0 );
			Debug.Assert( ( index + length ) <= this.Count );

			if( this.data64 == null || this.Count == 0 ) {
				return	-1;
			}

			return	this.PrivateBinarySearch( index, length, element );
		}

		public int RemoveDuplicates() {
			if( this.Count <= 1 ) {
				return 0;
			}
			int newIndex = 1, oldIndex;
			T lastUniqueElement = this[ 0 ];
			for( oldIndex = 1; oldIndex < this.Count; oldIndex ++ ) {
				T currentElement = this[ oldIndex ];
				if( currentElement != lastUniqueElement ) {
					this[ newIndex ++ ] = currentElement;
					lastUniqueElement = currentElement;
				}
			}
			this.count40 = newIndex;
			return	( oldIndex - newIndex );
		}

		//---------------------------------------------------------------------------------

		public void Sort() {
			if( this.data64 == null || this.Count <= 1 ) {
				return;
			}
			this.PrivateQuickSort( 0, this.Count - 1 );
		}

		private void PrivateQuickSort( int left, int right ) {
			// Can use the much faster jit helpers for array access.
			do {
				int i = left;
				int j = right;
				T x = this[ ( i + j ) >> 1 ];
				do {
					while( this[ i ] < x ) i++;
					while( x < this[ j ] ) j--;
					
					if( i > j ) {
						break;
					}
					if( i < j ) {
						// swap keys[i] and keys[j]
						T key = this[ i ];
						this[ i ] = this[ j ];
						this[ j ] = key;
					}
					i ++;
					j --;
				} while( i <= j );

				if( j - left <= right - i ) {
					if( left < j ) {
						this.PrivateQuickSort( left, j );
					}
					left = i;
				}
				else {
					if( i < right ) {
						this.PrivateQuickSort( i, right );
					}
					right = j;
				}
			} while( left < right );
		}
				
		//---------------------------------------------------------------------------------
	}

	//=====================================================================================
	//=====================================================================================

}
