// ==++==
// 
//   
//    Copyright (c) 2002 Microsoft Corporation.  All rights reserved.
//   
//    The use and distribution terms for this software are contained in the file
//    named license.txt, which can be found in the root of this distribution.
//    By using this software in any fashion, you are agreeing to be bound by the
//    terms of this license.
//   
//    You must not remove this notice, or any other, from this software.
//   
// 
// ==--==
/*=============================================================================
**
** Class: Queue
**
**                                                                              
**
** Purpose: A circular-array implementation of a queue.
**
** Date: October 4, 1999
**
=============================================================================*/

using System;

namespace Exocortex.Collections.Specialized {

	using T = System.UInt64;
	
	// A simple Queue of objects.  Internally it is implemented as a circular
    // buffer, so Enqueue can be O(n).  Dequeue is O(1).
	public class UInt64Queue {
        private T[] _array;
        private int _head;       // First valid element in the queue
        private int _tail;       // Last valid element in the queue
        private int _size;       // Number of elements.
        private int _growFactor; // 100 == 1.0, 130 == 1.3, 200 == 2.0
    	
        private const int _MinimumGrow = 4;
        private const int _ShrinkThreshold = 32;
    
        // Creates a queue with room for capacity objects. The default initial
        // capacity and grow factor are used.
        public UInt64Queue() 
            : this(32, (float)2.0) {
        }
    
        // Creates a queue with room for capacity objects. The default grow factor
        // is used.
        //
        public UInt64Queue(int capacity) 
            : this(capacity, (float)2.0) {
        }
        
        // Creates a queue with room for capacity objects. When full, the new
        // capacity is set to the old capacity * growFactor.
        //
        public UInt64Queue(int capacity, float growFactor) {
    		if (capacity < 0)
    			throw new ArgumentOutOfRangeException("capacity", "ArgumentOutOfRange_NeedNonNegNum" );
            if (!(growFactor >= 1.0 && growFactor <= 10.0))
                throw new ArgumentOutOfRangeException("growFactor", "ArgumentOutOfRange_QueueGrowFactor" );
    
            _array = new T[capacity];
            _head = 0;
            _tail = 0;
            _size = 0;
            _growFactor = (int)(growFactor * 100);
        }   
    
    	public int Count {
    		get { return _size; }
    	}
    	    	
        // Removes all Objects from the queue.
        public void Clear() {
            if (_head < _tail)
                Array.Clear(_array, _head, _size);
            else {
                Array.Clear(_array, _head, _array.Length - _head);
                Array.Clear(_array, 0, _tail);
            }
    
            _head = 0;
            _tail = 0;
            _size = 0;
        }
    
    	// CopyTo copies a collection into an Array, starting at a particular
    	// index into the array.
    	// 
    	public void CopyTo(Array array, int index)
    	{
    		if (array==null)
    			throw new ArgumentNullException("array");
            if (array.Rank != 1)
                throw new ArgumentException( "Arg_RankMultiDimNotSupported" );
    		if (index < 0)
    			throw new ArgumentOutOfRangeException("index", "ArgumentOutOfRange_Index" );
            int arrayLen = array.Length;
            if (arrayLen - index < _size)
                throw new ArgumentException( "Argument_InvalidOffLen" );
    
    		int numToCopy = (arrayLen-index < _size) ? arrayLen-index : _size;
            if (numToCopy==0)
                return;
    		if (_head < _tail) {
    			Array.Copy(_array, _head, array, index, numToCopy);
    		} else {
    			int firstPart = (_array.Length - _head < numToCopy) ? _array.Length - _head : numToCopy;
    			Array.Copy(_array, _head, array, index, firstPart);
    			numToCopy -= firstPart;
    			Array.Copy(_array, 0, array, index+_array.Length - _head, numToCopy);
    		}
    	}
    	
        // Adds obj to the tail of the queue.
        //
        public void Enqueue(T element ) {
            if (_size == _array.Length) {
                int newcapacity = (int)((long)_array.Length * (long)_growFactor / 100);
                if (newcapacity < _array.Length + _MinimumGrow) {
                    newcapacity = _array.Length + _MinimumGrow;
                }
                SetCapacity(newcapacity);
            }
    
            _array[_tail] = element;
            _tail = (_tail + 1) % _array.Length;
            _size++;
        }
    
        // Removes the object at the head of the queue and returns it. If the queue
        // is empty, this method simply returns null.
        public T Dequeue() {
            if (_size == 0)
                throw new InvalidOperationException( "InvalidOperation_EmptyQueue" );
    
            T removed = _array[_head];
            _array[_head] = (T) 0;
            _head = (_head + 1) % _array.Length;
            _size--;
            return removed;
        }
    
        // Returns the object at the head of the queue. The object remains in the
        // queue. If the queue is empty, this method throws an 
        // InvalidOperationException.
        /// <include file='doc\Queue.uex' path='docs/doc[@for="Queue.Peek"]/*' />
        public T Peek() {
            if (_size == 0)
                throw new InvalidOperationException( "InvalidOperation_EmptyQueue" );
    
            return _array[_head];
        }
    
		// Returns true if the queue contains at least one object equal to obj.
        // Equality is determined using obj.Equals().
        //
        // Exceptions: ArgumentNullException if obj == null.
		public bool Contains(T element) {
    		int index = _head;
            int count = _size;
    
            while (count-- > 0) {
                if( _array[ index ] == element ) {
					return true;
                }
                index = (index + 1) % _array.Length;
            }
    
            return false;
        }
    
        internal T GetElement(int i)
        {
            return _array[(_head + i) % _array.Length];
        }
    
        // Iterates over the objects in the queue, returning an array of the
        // objects in the Queue, or an empty array if the queue is empty.
        // The order of elements in the array is first in to last in, the same
        // order produced by successive calls to Dequeue.
    	/// <include file='doc\Queue.uex' path='docs/doc[@for="Queue.ToArray"]/*' />
    	public T[] ToArray()
    	{
    		T[] arr = new T[_size];
            if (_size==0)
                return arr;
    
            if (_head < _tail) {
                Array.Copy(_array, _head, arr, 0, _size);
            } else {
                Array.Copy(_array, _head, arr, 0, _array.Length - _head);
                Array.Copy(_array, 0, arr, _array.Length - _head, _tail);
            }
    
    		return arr;
    	}
    
    
        // PRIVATE Grows or shrinks the buffer to hold capacity objects. Capacity
        // must be >= _size.
        private void SetCapacity(int capacity) {
            T[] newarray = new T[capacity];
            if (_size > 0) {
                if (_head < _tail) {
                    Array.Copy(_array, _head, newarray, 0, _size);
                } else {
                    Array.Copy(_array, _head, newarray, 0, _array.Length - _head);
                    Array.Copy(_array, 0, newarray, _array.Length - _head, _tail);
                }
            }
    
            _array = newarray;
            _head = 0;
            _tail = _size;
        }
    
        public void TrimToSize() {
            SetCapacity(_size);
        }
    
    }
}
