/*
 * BSD Licence:
 * Copyright (c) 2001, Ben Houston [ ben@exocortex.org ]
 * Exocortex Technologies [ www.exocortex.org ]
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright 
 * notice, this list of conditions and the following disclaimer in the 
 * documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the <ORGANIZATION> nor the names of its contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

using System;
using System.Diagnostics;
using System.Collections;
using System.Threading;
using System.Reflection;
using System.Text;

using Exocortex.Diagnostics;

namespace Exocortex {

	public class AssertionFailedException : Exception {
		public AssertionFailedException() {}
		public AssertionFailedException( string message ) : base( message ) {
		}
	}

	public class Debug2 {

		//==================================================================

		// Prevent from begin created
		private Debug2() {
		}

		//==================================================================

		[Conditional("DEBUG")]
		static public void Assert( bool condition ) {
			if( condition == false ) {
				throw new AssertionFailedException();
			}
		}

		[Conditional("DEBUG")]
		static public void Assert( bool condition, string message ) {
			if( condition == false ) {
				throw new AssertionFailedException( message );
			}
		}

		//==================================================================

		//[Conditional("DEBUG")]
		static public string GetCallingMethod( int levels ) {
			StackTrace st = new StackTrace( 2, true );
			StackFrame frame = st.GetFrame( levels );
			MethodBase method = frame.GetMethod();
			// C:\Development\ExoEngine\Geometry\SkyBox.cs(36):
			return	method.DeclaringType.Name + "." + method.Name;
		}

		[Conditional("DEBUG")]
		static public void TraceCallingFunction() {
			StackTrace st = new StackTrace( 2, true );
			StackFrame frame = st.GetFrame( 0 );
			MethodBase method = frame.GetMethod();
			// C:\Development\ExoEngine\Geometry\SkyBox.cs(36):
			Debug.WriteLine( "[function trace] " + method.DeclaringType + "." + method.Name );
			Debug.WriteLine( frame.GetFileName() + "(" + frame.GetFileLineNumber() + ")" );
		}

		[Conditional("DEBUG")]
		static public void TraceFunction() {
			StackTrace st = new StackTrace( 1, true );
			StackFrame frame = st.GetFrame( 0 );
			MethodBase method = frame.GetMethod();
			// C:\Development\ExoEngine\Geometry\SkyBox.cs(36):
			Debug.WriteLine( "[function trace] " + method.DeclaringType + "." + method.Name );
			Debug.WriteLine( frame.GetFileName() + "(" + frame.GetFileLineNumber() + ")" );
		}

		[Conditional("DEBUG")]
		static public void TraceFunction( string id ) {
			StackTrace st = new StackTrace( 1, true );
			StackFrame frame = st.GetFrame( 0 );
			MethodBase method = frame.GetMethod();
			// C:\Development\ExoEngine\Geometry\SkyBox.cs(36):
			Debug.WriteLine( "[function trace:" + id + "] " + method.DeclaringType + "." + method.Name );
			Debug.WriteLine( frame.GetFileName() + "(" + frame.GetFileLineNumber() + ")" );
		}

		[Conditional("DEBUG")]
		static public void TraceFunction( int id ) {
			StackTrace st = new StackTrace( 1, true );
			StackFrame frame = st.GetFrame( 0 );
			MethodBase method = frame.GetMethod();
			// C:\Development\ExoEngine\Geometry\SkyBox.cs(36):
			Debug.WriteLine( "[function trace:" + id + "] " + method.DeclaringType + "." + method.Name );
			Debug.WriteLine( frame.GetFileName() + "(" + frame.GetFileLineNumber() + ")" );
		}

		[Conditional("DEBUG")]
		static public void TraceThread() {
			Thread currentThread = Thread.CurrentThread;
			if( currentThread.Name == null || currentThread.Name.Length == 0 ) {
				currentThread.Name = new Random().Next( 1000 ).ToString();
			}
			Debug.WriteLine( "[thread trace] CurrentThread.Name = '" + currentThread.Name + "'" );
			//Debug.WriteLine( frame.GetFileName() + "(" + frame.GetFileLineNumber() + ")" );
		}

		[Conditional("DEBUG")]
		static public void TraceStack() {
			Debug.WriteLine( "[stack trace]" );
			StackTrace st = new StackTrace( 1, true );
			for( int i = 0; i < st.FrameCount; i ++ ) {
				StackFrame frame = st.GetFrame( i );
				MethodBase method = frame.GetMethod();
				Debug.WriteLine( "   " + method.DeclaringType + "." + method.Name + "   " + frame.GetFileName() + " line " + frame.GetFileLineNumber() );
			}
		}

		[Conditional("DEBUG")]
		static public void TraceStack( int maximumDepth ) {
			Debug.WriteLine( "[stack trace (maximumDepth=" + maximumDepth + ")]" );
			StackTrace st = new StackTrace( 1, true );
			for( int i = 0; i < Math.Min( st.FrameCount, maximumDepth ); i ++ ) {
				StackFrame frame = st.GetFrame( i );
				MethodBase method = frame.GetMethod();
				Debug.WriteLine( "   " + method.DeclaringType + "." + method.Name + "   " + frame.GetFileName() + " line " + frame.GetFileLineNumber() );
			}
		}

		[Conditional("DEBUG")]
		static public void	WriteVar( string variableName, object variable ) {
			StringBuilder stringBuilder = new StringBuilder( 50 );
			
			stringBuilder.Append( variableName );
			stringBuilder.Append( '[' );
			if( variable != null ) {
				stringBuilder.Append( variable.GetType().Name );
			}
			else {
				stringBuilder.Append( "(unknown)" );
			}
			stringBuilder.Append( "] = " );
			if( variable != null ) {
				stringBuilder.Append( variable.ToString() );
			}
			else {
				stringBuilder.Append( "(null)" );
			}

			Debug.WriteLine( stringBuilder.ToString() );
		}

		[Conditional("DEBUG")]
		static public void	WriteArray( string arrayName, Array array ) {
			StringBuilder stringBuilder = new StringBuilder( 100 );
			stringBuilder.Append( arrayName );
			stringBuilder.Append( '[' );
			if( array != null ) {
				stringBuilder.Append( array.GetType().GetElementType().Name );
			}
			else {
				stringBuilder.Append( "(unknown)" );
			}
			stringBuilder.Append( "] = " );
			if( array != null ) {
				stringBuilder.Append( "( " );
				for( int i = 0; i < array.Length; i ++ ) {
					stringBuilder.Append( array.GetValue( i ).ToString() );
					stringBuilder.Append( ' ' );
				}
				stringBuilder.Append( ')' );				
			}
			else {
				stringBuilder.Append( "(null)" );
			}
		
			Debug.WriteLine( stringBuilder.ToString() );
		}

		[Conditional("DEBUG")]
		static public void Collect() {
			Debug2.Push();
			System.GC.Collect();
			Debug2.Pop();
		}
			
		[Conditional("DEBUG")]
		static public void WriteSeparator() {
			Debug.WriteLine( "-------------------------------------------------------------------------------" );
		}
		[Conditional("DEBUG")]
		static public void WriteLine() {
			Debug.WriteLine( "" );
		}
		[Conditional("DEBUG")]
		static public void WriteLine( string s ) {
			Debug.WriteLine( s );
		}
		[Conditional("DEBUG")]
		static public void WriteLines( string[] ss ) {
			foreach( string s in ss ) {
				Debug.WriteLine( s );
			}
		}
		[Conditional("DEBUG")]
		static public void Write( string s ) {
			Debug.Write( s );
		}
		[Conditional("DEBUG")]
		static public void Writes( string[] ss ) {
			foreach( string s in ss ) {
				Debug.Write( s );
			}
		}

		//==================================================================

		[Conditional("DEBUG")]
		static public void Push() {
			StackTrace st = new StackTrace( 1, true );
			StackFrame frame = st.GetFrame( 0 );
			MethodBase method = frame.GetMethod();

			string strFrameName = method.DeclaringType + "." + method.Name;
			
			Debug.WriteLine( ProfilingStack.Push( strFrameName ) );
			Debug.Indent();
		}

		[Conditional("DEBUG")]
		static public void Push( String sectionName ) {
			Debug.WriteLine( ProfilingStack.Push( sectionName ) );
			Debug.Indent();
		}

		[Conditional("DEBUG")]
		static public void Pop() {
			Debug.Unindent();
			Debug.WriteLine( ProfilingStack.Pop() );
		}

		//==================================================================

	}
}
