/*
 * BSD Licence:
 * Copyright (c) 2001, Ben Houston [ ben@exocortex.org ]
 * Exocortex Technologies [ www.exocortex.org ]
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright 
 * notice, this list of conditions and the following disclaimer in the 
 * documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the <ORGANIZATION> nor the names of its contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

using System;
using System.Diagnostics;
using System.Drawing;
using System.Collections;
using System.ComponentModel;
using System.IO;
using System.Windows.Forms;
using System.Reflection;
using System.Resources;
using System.Data;
using System.Threading;

namespace Exocortex.Diagnostics
{
	/// <summary>
	/// Summary description for ErrorLog.
	/// </summary>
	public class BugTracking {
		
		//--------------------------------------------------------------------------------------------

		static public void OnThreadException( object o, ThreadExceptionEventArgs e ) {
			BugTracking.LogException( e.Exception );
		}

		static public void LogException( Exception e ) {
			string fileName = BugTracking.WriteExceptionRecord( e );
			System.Windows.Forms.MessageBox.Show( null,
				"A fatal error occured.  Please send the following bug report to bugs@exocortex.org:\n" + 
				fileName,
				Application.ProductName,
				MessageBoxButtons.OK,
				MessageBoxIcon.Error );

			System.Diagnostics.Process proc = new System.Diagnostics.Process();
			proc.EnableRaisingEvents = false;
			proc.StartInfo.FileName = "notepad";
			proc.StartInfo.Arguments = fileName;
			proc.Start();
		}

		//--------------------------------------------------------------------------------------------

		protected class LogEntry {
			public string		Text;
			public DateTime	DateTime;
			public LogEntry( string text ) {
				this.Text		= text;
				this.DateTime	= DateTime.Now;
			}
			public override string	ToString() {
				return	"[" + this.DateTime.ToLocalTime() + "] " + this.Text;
			}
		}

		static protected ArrayList	s_logEntries	= new ArrayList();

		static public void Log( string text ) {
			s_logEntries.Add( new LogEntry( text ) );
		}

		//--------------------------------------------------------------------------------------------

		static protected Hashtable	s_htWatches = new Hashtable();

		static public void	SetWatch( string name, object value ) {
			Debug.Assert( name != null );
			if( s_htWatches.ContainsKey( name ) ) {
				s_htWatches.Remove( name );
			}

			s_htWatches.Add( name, ( value != null ) ? new WeakReference( value ) : null );
			Debug.Assert( s_htWatches.ContainsKey( name ) == true );
		}

		static public void	RemoveWatch( string name ) {
			Debug.Assert( name != null );
			Debug.Assert( s_htWatches.ContainsKey( name ) == true );
			
			s_htWatches.Remove( name );
		}

		static public string	GetWatchString( string name ) {
			Debug.Assert( name != null );
			Debug.Assert( s_htWatches.ContainsKey( name ) == true );
			
			WeakReference wr = (WeakReference) s_htWatches[ name ];
			if( wr != null ) {
				if( wr.IsAlive == true ) {
					object value = wr.Target;
					return	value.ToString() + " [" + value.GetType().Name + "]";
				}
				else {
					return	"(garbageCollected)";
				}
			}
			else {
				return	"(null)";
			}
		}

		//--------------------------------------------------------------------------------------------
		
		static protected string	s_outputDirectory = "";
		static public string	OutputDirectory {
			get	{	return	s_outputDirectory;	}
			set	{	s_outputDirectory = value;	}
		}

		static protected string	s_applicationName = "";
		static public string	ApplicationName {
			get	{	return	s_applicationName;	}
			set	{	s_applicationName = value;	}
		}
		
		//--------------------------------------------------------------------------------------------

		static public string WriteExceptionRecord( Exception e ) {
			try {
				string directory = OutputDirectory;
				if( directory == "" ) {
					directory = Application.StartupPath;
				}
				if( Directory.Exists( directory ) == false ) {
					Directory.CreateDirectory( directory );
				}
				string	fileName = Path.Combine( directory, s_applicationName + "_error_" + String2.GetSimpleDateTime() + "_" + e.GetType().Name + ".log" );
				StreamWriter sw = new StreamWriter( fileName, true );
			
				sw.WriteLine( "ErrorLog.WriteExceptionLog() >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>" );

				try {
					DateTime now = DateTime.Now;
					sw.WriteLine( "Date Stamp" );
					sw.WriteLine( "  CurrentDate: " + now.ToLongDateString() );
					sw.WriteLine( "  CurrentTime: " + now.ToLongTimeString() );
				}
				catch( Exception ex ) {
					sw.WriteLine( "<error-CurrentDateTime>" + ex.ToString() );
				}
			
				sw.WriteLine();
				sw.Flush();
						
				try {
					sw.WriteLine( "Exception Details" );
					sw.WriteLine( "  Exception.Type: " + e.GetType().Name );
					sw.WriteLine( "  Exception.Message: " + e.Message );
					sw.WriteLine( "  Exception.StackTrace: " );
					sw.WriteLine( e.StackTrace );
					sw.WriteLine( "  Exception.Source: " + e.Source );
					sw.WriteLine( "  Exception.HelpLink: " + e.HelpLink );
					sw.WriteLine( "  Exception.TargetSite: " + e.TargetSite );
				}
				catch( Exception ex ) {
					sw.WriteLine( "<error-Exception>" + ex.ToString() );
				}
			
				sw.WriteLine();
				sw.Flush();
			
				try {
					sw.WriteLine( "ErrorTracking Watches" );
					foreach( string watchName in s_htWatches.Keys ) {
						sw.WriteLine( "  Watch[" + watchName + "]: " + GetWatchString( watchName ) );
					}
				}
				catch( Exception ex ) {
					sw.WriteLine( "<error-Watches>" + ex.ToString() );
				}
		
				sw.WriteLine();
				sw.Flush();
			
				try {
					sw.WriteLine( "ErrorTracking Log" );
					foreach( LogEntry logEntry in s_logEntries ) {
						sw.WriteLine( "  LogEntry" + logEntry.ToString() );
					}
					sw.Flush();
				}
				catch( Exception ex ) {
					sw.WriteLine( "<error-LogEntry>" + ex.ToString() );
				}
				
				sw.WriteLine();
				sw.Flush();
						
				try {
					sw.WriteLine( "Memory and CPU Stats" );
					sw.WriteLine( "  GC.TotalMemory: " + GC.GetTotalMemory( false ) );
					sw.WriteLine( "  Environment.WorkingSet: " + Environment.WorkingSet );
				}
				catch( Exception ex ) {
					sw.WriteLine( "<error-GC>" + ex.ToString() );
				}
				
				sw.WriteLine();
				sw.Flush();
						
				try {
					try {
						sw.WriteLine( "Application Information" );
						sw.WriteLine( "  Application.ExecutablePath: " + Application.ExecutablePath );
						sw.WriteLine( "  Application.CurrentDirectory: " + Directory.GetCurrentDirectory() );
						sw.WriteLine( "  Application.StartupPath: " + Application.StartupPath );
						sw.WriteLine( "  Application.ProductName: " + Application.ProductName );
						sw.WriteLine( "  Application.ProductVersion: " + Application.ProductVersion );
					}
					catch( BadImageFormatException ) {
					}
				}
				catch( Exception ex ) {
					if( !( ex is BadImageFormatException ) )
						sw.WriteLine( "<error-Application>" + ex.ToString() );
				}
			
				sw.WriteLine();
				sw.Flush();
						
				try {
					sw.WriteLine( "Assembly Information (Executing)" );
					Assembly execAssembly = Assembly.GetExecutingAssembly();
					sw.WriteLine( "  ExecutingAssembly.CodeBase: " + execAssembly.CodeBase );
					sw.WriteLine( "  ExecutingAssembly.Location: " + execAssembly.Location );
					sw.WriteLine( "  ExecutingAssembly.GlobalAssemblyCache: " + execAssembly.GlobalAssemblyCache );
				}
				catch( Exception ex ) {
					sw.WriteLine( "<error-ExecutingAssembly>" + ex.ToString() );
				}
			
				sw.WriteLine();
				sw.Flush();
						
					
				try {
					Assembly selfAssembly = Assembly.GetAssembly( typeof( BugTracking ) );
					sw.WriteLine( "Assembly Information (Current)" );
					sw.WriteLine( "  CurrentAssembly.CodeBase: " + selfAssembly.CodeBase );
					sw.WriteLine( "  CurrentAssembly.Location: " + selfAssembly.Location );
					sw.WriteLine( "  CurrentAssembly.GlobalAssemblyCache: " + selfAssembly.GlobalAssemblyCache );
				}
				catch( Exception ex ) {
					sw.WriteLine( "<error-CurrentAssembly>" + ex.ToString() );
				}
				
				sw.WriteLine();
				sw.Flush();
						
				try {
					Thread thread = Thread.CurrentThread;
					sw.WriteLine( "Thread Information" );
					sw.WriteLine( "  CurrentThread.Name: " + thread.Name );
					sw.WriteLine( "  CurrentThread.Priority: " + thread.Priority );
				}
				catch( Exception ex ) {
					sw.WriteLine( "<error-CurrentThread>" + ex.ToString() );
				}
			
				sw.WriteLine();
				sw.Flush();
						
				try {
					sw.WriteLine( "Process Information" );
					Process process = Process.GetCurrentProcess();
					sw.WriteLine( "  CurrentProcess.Name: " + process.ProcessName );
					sw.WriteLine( "  CurrentProcess.MachineName: " + process.MachineName );
					sw.WriteLine( "  CurrentProcess.MainModule: " + process.MainModule );
					sw.WriteLine( "  CurrentProcess.StartDate: " + process.StartTime.ToLongDateString() );
					sw.WriteLine( "  CurrentProcess.StartTime: " + process.StartTime.ToLongTimeString() );
					sw.WriteLine( "  CurrentProcess.UserProcessorTime: " + process.UserProcessorTime );
					sw.WriteLine( "  CurrentProcess.TotalProcessorTime: " + process.TotalProcessorTime );
				}
				catch( Exception ex ) {
					sw.WriteLine( "<error-CurrentProcess>" + ex.ToString() );
				}
			
				sw.WriteLine();
				sw.Flush();
						
				try {
					OperatingSystem os = Environment.OSVersion;
					sw.WriteLine( "Operating System Information" );
					sw.WriteLine( "  Environment.OSVersion.Platform: " + os.Platform );
					sw.WriteLine( "  Environment.OSVersion.Version: " + os.Version );
					sw.WriteLine();
			
					Version ver = Environment.Version;
					sw.WriteLine( ".NET Platform Information" );
					sw.WriteLine( "  Environment.Version.Major: " + ver.Major );
					sw.WriteLine( "  Environment.Version.Minor: " + ver.Minor );
					sw.WriteLine( "  Environment.Version.Revision: " + ver.Revision );
					sw.WriteLine( "  Environment.Version.Build: " + ver.Build );
					sw.WriteLine();
			
					sw.WriteLine( "Misc Environment Information" );
					sw.WriteLine( "  Environment.MachineName: " + Environment.MachineName );
					sw.WriteLine( "  Environment.UserName: " + Environment.UserName );
					sw.WriteLine( "  Environment.SystemDirectory: " + Environment.SystemDirectory );
					sw.WriteLine( "  Environment.TickCount: " + Environment.TickCount );
					sw.WriteLine();

					sw.WriteLine( "Command Line" );
					sw.WriteLine( "  Environment.CommandLine: " + Environment.CommandLine );
					string[] args = Environment.GetCommandLineArgs();
					if( args != null ) {
						for( int i = 0; i < args.Length; i ++ ) {
							sw.WriteLine( "  Environment.CommandLineArgs[" + i + "]: " + args[i] );
						}
					}
					sw.WriteLine();
					
					sw.WriteLine( "Current Call Stack" );
					sw.WriteLine( "  Environment.StackTrace: " );
					sw.WriteLine( Environment.StackTrace );
				}
				catch( Exception ex ) {
					sw.WriteLine( "<error-Environment>" + ex.ToString() );
				}
			
				sw.WriteLine();
				sw.Flush();
				sw.Close();
				
				return	fileName;
			}
			catch( Exception ) {
				return	"-- error writing log file --";
			}
		}
	}
}
