/*
 * BSD Licence:
 * Copyright (c) 2001, Ben Houston [ ben@exocortex.org ]
 * Exocortex Technologies [ www.exocortex.org ]
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright 
 * notice, this list of conditions and the following disclaimer in the 
 * documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the <ORGANIZATION> nor the names of its contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

using System;

namespace Exocortex.Drawing.Drawing3D
{
	/// <summary>
	/// Summary description for Size3F.
	/// </summary>
	public struct Size3F {

		//----------------------------------------------------------------------------
		
		private float _width;
		private float _height;
		private float _depth;

		//----------------------------------------------------------------------------
		
		public Size3F( float width, float height, float depth ) {
			_width = width;	
			_height = height;
			_depth = depth;	
		}

		//----------------------------------------------------------------------------
		
		public float Width {
			get {	return	_width;	}
			set	{
				_width = value;	
			}
		}

		public float Height {
			get {	return	_height;	}
			set	{
				_height = value;
			}
		}
	
		public float Depth {
			get {	return	_depth;	}
			set	{
				_depth = value;	
			}
		}

		//----------------------------------------------------------------------------

		public bool	IsPositive {
			get {
				return	( _width >= 0 ) && ( _height >= 0 ) && ( _depth >= 0 );
			}
		}
		
		//----------------------------------------------------------------------------

		public float	GetVolume() {
			return	_width * _depth * _height;
		}
		
		//----------------------------------------------------------------------------
		
		public float	this[ int dimension ] {
			get {
				switch( dimension ) {
				case 0:
					return	this.Width;
				case 1:
					return	this.Height;
				case 2:
					return	this.Depth;
				default:
					throw new ArgumentOutOfRangeException( "dimension", dimension, "Dimension must be 0, 1 or 2" );
				}
			}
			set {
				switch( dimension ) {
				case 0:
					this.Width = value;
					break;
				case 1:
					this.Height = value;
					break;
				case 2:
					this.Depth = value;
					break;
				default:
					throw new ArgumentOutOfRangeException( "dimension", dimension, "Dimension must be 0, 1 or 2" );
				}
			}

		}

		//----------------------------------------------------------------------------

		public override string ToString() {
			return	"[ " + this.Width + ", " + this.Height + ", " + this.Depth + " ]";
		}

		//-----------------------------------------------------------------------------------------

		public override bool Equals( object o ) {
			if( o is Size3F ) {
				Size3F size = (Size3F) o;
				return	( this.Width == size.Width ) && ( this.Height == size.Height ) && ( this.Depth == size.Depth );
			}
			return	false;
		}

		public override int	GetHashCode() {
			return	( this.Width.GetHashCode() ^ this.Height.GetHashCode() ^ this.Depth.GetHashCode() );
		}
		
		//-----------------------------------------------------------------------------------------
		
		public static bool operator==( Size3F a, Size3F b ) {
			return	( a.Width == b.Width ) && ( a.Height == b.Height ) && ( a.Depth == b.Depth );
		}
		public static bool operator!=( Size3F a, Size3F b ) {
			return	( a.Width != b.Width ) || ( a.Height != b.Height ) || ( a.Depth != b.Depth );
		}

		public static Size3F operator+( Size3F a, Size3F b ) {
			return	new Size3F( a.Width + b.Width, a.Height + b.Height, a.Depth + b.Depth );
		}
		public static Size3F operator-( Size3F a, Size3F b ) {
			return	new Size3F( a.Width - b.Width, a.Height - b.Height, a.Depth - b.Depth );
		}
		public static Size3F operator*( Size3F a, float f ) {
			return	new Size3F( (int)( a.Width * f ), (int)( a.Height * f ), (int)( a.Depth * f ) );
		}
		public static Size3F operator/( Size3F a, float f ) {
			if( f == 0 ) {
				throw new DivideByZeroException( "parameter 'f' can not be zero" );
			}
			return	a * ( 1 / f );
		}

		//----------------------------------------------------------------------------

		static public readonly	Size3F	Empty	=	new Size3F( 0, 0, 0 );
		static public readonly	Size3F	Unit	=	new Size3F( 1, 1, 1 );

		//----------------------------------------------------------------------------

		static public Size3F	Max( Size3F a, Size3F b ) {
			return new Size3F(
				Math.Max( a.Width, b.Width ),
				Math.Max( a.Height, b.Height ),
				Math.Max( a.Depth, b.Depth ) );
		}
	
		static public Size3F	Min( Size3F a, Size3F b ) {
			return new Size3F(
				Math.Min( a.Width, b.Width ),
				Math.Min( a.Height, b.Height ),
				Math.Min( a.Depth, b.Depth ) );
		}

		//----------------------------------------------------------------------------

	}
}
