using System;
using System.Text;
using System.Diagnostics;

namespace Exocortex.LowLevel
{
	/// <summary>
	/// Summary description for BitUtils.
	/// </summary>
	public class BitUtils
	{
		private BitUtils() {}

		//---------------------------------------------------------------------------------------

		static public ushort	ClearBitUInt16( ushort value, int bit ) {
			Debug.Assert( 0 <= bit && bit < 16 );
			unchecked {
				return	(ushort)( value & ~( 1 << bit ) );
			}
		}

		static public int		ClearBitInt32( int value, int bit ) {
			Debug.Assert( 0 <= bit && bit < 32 );
			unchecked {
				return	( value & ~( 1 << bit ) );
			}
		}

		static public uint		ClearBitUInt32( uint value, int bit ) {
			Debug.Assert( 0 <= bit && bit < 32 );
			unchecked {
				return	( value & ~( 1u << bit ) );
			}
		}

		static public ulong		ClearBitUInt64( ulong value, int bit ) {
			Debug.Assert( 0 <= bit && bit < 64 );
			unchecked {
				return	( value & ~( 1ul << bit ) );
			}
		}
		//---------------------------------------------------------------------------------------

		static public ushort	SetBitUInt16( ushort value, int bit ) {
			Debug.Assert( 0 <= bit && bit < 16 );
			unchecked {
				return	(ushort)( value | ( 1 << bit ) );
			}
		}

		static public int		SetBitInt32( int value, int bit ) {
			Debug.Assert( 0 <= bit && bit < 32 );
			unchecked {
				return	( value | ( 1 << bit ) );
			}
		}

		static public uint		SetBitUInt32( uint value, int bit ) {
			Debug.Assert( 0 <= bit && bit < 32 );
			unchecked {
				return	( value | ( 1u << bit ) );
			}
		}

		static public ulong		SetBitUInt64( ulong value, int bit ) {
			Debug.Assert( 0 <= bit && bit < 64 );
			unchecked {
				return	( value | ( 1ul << bit ) );
			}
		}

		//---------------------------------------------------------------------------------------

		static public ulong		CreateBitMaskUInt32( int numberOfBits ) {
			Debug.Assert( 0 <= numberOfBits && numberOfBits < 32 );
			uint	bitMask = 0;
			for( int bit = 0; bit < numberOfBits; bit ++ ) {
				bitMask = SetBitUInt32( bitMask, bit );
			}
			return	bitMask;
		}

		static public ulong		CreateBitMaskUInt64( int numberOfBits ) {
			Debug.Assert( 0 <= numberOfBits && numberOfBits < 64 );
			ulong	bitMask = 0;
			for( int bit = 0; bit < numberOfBits; bit ++ ) {
				bitMask = SetBitUInt64( bitMask, bit );
			}
			return	bitMask;
		}

		//---------------------------------------------------------------------------------------

		static public bool	IsBitSetUInt16( ushort value, int bit ) {
			Debug.Assert( 0 <= bit && bit < 16 );
			unchecked {
				return	( value & ( 1 << bit ) ) != 0;
			}
		}

		static public bool	IsBitSetInt32( int value, int bit ) {
			Debug.Assert( 0 <= bit && bit < 32 );
			unchecked {
				return	( value & ( 1u << bit ) ) != 0;
			}
		}

		static public bool	IsBitSetUInt32( uint value, int bit ) {
			Debug.Assert( 0 <= bit && bit < 32 );
			unchecked {
				return	( value & ( 1u << bit ) ) != 0;
			}
		}

		static public bool	IsBitSetUInt64( ulong value, int bit ) {
			Debug.Assert( 0 <= bit && bit < 64 );
			unchecked {
				return	( value & ( 1ul << bit ) ) != 0;
			}
		}
		
		//---------------------------------------------------------------------------------------

		static public string	GetBitStringUInt16( ushort value ) {
			StringBuilder bitString = new StringBuilder();
			for( int bit = 15; bit >= 0; bit -- ) {
				if( BitUtils.IsBitSetUInt16( value, bit ) ) {
					bitString.Append( '1' );
				}
				else {
					bitString.Append( '0' );
				}
			}
			return	bitString.ToString();
		}

		static public string	GetBitStringUInt32( uint value ) {
			StringBuilder bitString = new StringBuilder();
			for( int bit = 31; bit >= 0; bit -- ) {
				if( BitUtils.IsBitSetUInt32( value, bit ) ) {
					bitString.Append( '1' );
				}
				else {
					bitString.Append( '0' );
				}
			}
			return	bitString.ToString();
		}

		static public string	GetBitStringUInt64( ulong value ) {
			StringBuilder bitString = new StringBuilder();
			for( int bit = 63; bit >= 0; bit -- ) {
				if( BitUtils.IsBitSetUInt64( value, bit ) ) {
					bitString.Append( '1' );
				}
				else {
					bitString.Append( '0' );
				}
			}
			return	bitString.ToString();
		}

		//---------------------------------------------------------------------------------------

		static public int	CountBitsUInt16( ushort value ) {
			int count = 0;
			for( int bit = 0; bit < 16; bit ++ ) {
				if( BitUtils.IsBitSetUInt16( value, bit ) ) {
					count ++;
				}
			}
			return	count;
		}

		static public int	CountBitsInt32( int value ) {
			int count = 0;
			for( int bit = 0; bit < 32; bit ++ ) {
				if( BitUtils.IsBitSetInt32( value, bit ) ) {
					count ++;
				}
			}
			return	count;
		}

		static public int	CountBitsUInt32( uint value ) {
			int count = 0;
			for( int bit = 0; bit < 32; bit ++ ) {
				if( BitUtils.IsBitSetUInt32( value, bit ) ) {
					count ++;
				}
			}
			return	count;
		}

		static public int	CountBitsUInt64( ulong value ) {
			int count = 0;
			for( int bit = 0; bit < 64; bit ++ ) {
				if( BitUtils.IsBitSetUInt64( value, bit ) ) {
					count ++;
				}
			}
			return	count;
		}
		
		//---------------------------------------------------------------------------------------
	
		static public int[]		GetBitIndicesUInt16( ushort value ) {
			int totalCount = BitUtils.CountBitsUInt16( value );
			int[] bitIndices = new int[ totalCount ];
			int currentCount = 0;
			for( int bit = 0; bit < 16; bit ++ ) {
				if( BitUtils.IsBitSetUInt16( value, bit ) ) {
					bitIndices[ currentCount ++ ] = bit;
				}
			}
			Debug.Assert( currentCount == totalCount );
			return	bitIndices;
		}

		static public int[]		GetBitIndicesInt32( int value ) {
			int totalCount = BitUtils.CountBitsInt32( value );
			int[] bitIndices = new int[ totalCount ];
			int currentCount = 0;
			for( int bit = 0; bit < 32; bit ++ ) {
				if( BitUtils.IsBitSetInt32( value, bit ) ) {
					bitIndices[ currentCount ++ ] = bit;
				}
			}
			Debug.Assert( currentCount == totalCount );
			return	bitIndices;
		}

		static public int[]		GetBitIndicesUInt32( uint value ) {
			int totalCount = BitUtils.CountBitsUInt32( value );
			int[] bitIndices = new int[ totalCount ];
			int currentCount = 0;
			for( int bit = 0; bit < 32; bit ++ ) {
				if( BitUtils.IsBitSetUInt32( value, bit ) ) {
					bitIndices[ currentCount ++ ] = bit;
				}
			}
			Debug.Assert( currentCount == totalCount );
			return	bitIndices;
		}
		
		static public int[]		GetBitIndicesUInt64( ulong value ) {
			int totalCount = BitUtils.CountBitsUInt64( value );
			int[] bitIndices = new int[ totalCount ];
			int currentCount = 0;
			for( int bit = 0; bit < 64; bit ++ ) {
				if( BitUtils.IsBitSetUInt64( value, bit ) ) {
					bitIndices[ currentCount ++ ] = bit;
				}
			}
			Debug.Assert( currentCount == totalCount );
			return	bitIndices;
		}

		//---------------------------------------------------------------------------------------

		static public ushort	SetBitsUInt16( int[] bitIndices ) {
			Debug.Assert( bitIndices != null );
			Debug.Assert( bitIndices.Length <= 16 );
			ushort value = 0;
			for( int i = 0; i < bitIndices.Length; i ++ ) {
				value = BitUtils.SetBitUInt16( value, bitIndices[ i ] );
			}
			return	value;
		}

		static public uint		SetBitsUInt32( int[] bitIndices ) {
			Debug.Assert( bitIndices != null );
			Debug.Assert( bitIndices.Length <= 32 );
			uint value = 0;
			for( int i = 0; i < bitIndices.Length; i ++ ) {
				value = BitUtils.SetBitUInt32( value, bitIndices[ i ] );
			}
			return	value;
		}

		static public ulong		SetBitsUInt64( int[] bitIndices ) {
			Debug.Assert( bitIndices != null );
			Debug.Assert( bitIndices.Length <= 64 );
			ulong value = 0;
			for( int i = 0; i < bitIndices.Length; i ++ ) {
				value = BitUtils.SetBitUInt64( value, bitIndices[ i ] );
			}
			return	value;
		}

		//---------------------------------------------------------------------------------------

		static public int	LowestBitUInt16( ushort value ) {
			for( int lowestBit = 0; lowestBit < 16; lowestBit ++ ) {
				if( BitUtils.IsBitSetUInt16( value, lowestBit ) ) {
					return	lowestBit;
				}
			}
			return	-1;
		}

		static public int	LowestBitInt32( int value ) {
			for( int lowestBit = 0; lowestBit < 32; lowestBit ++ ) {
				if( BitUtils.IsBitSetInt32( value, lowestBit ) ) {
					return	lowestBit;
				}
			}
			return	-1;
		}

		static public int	LowestBitUInt32( uint value ) {
			for( int lowestBit = 0; lowestBit < 32; lowestBit ++ ) {
				if( BitUtils.IsBitSetUInt32( value, lowestBit ) ) {
					return	lowestBit;
				}
			}
			return	-1;
		}
		
		static public int	LowestBitUInt64( ulong value ) {
			for( int lowestBit = 0; lowestBit < 64; lowestBit ++ ) {
				if( BitUtils.IsBitSetUInt64( value, lowestBit ) ) {
					return	lowestBit;
				}
			}
			return	-1;
		}

		//---------------------------------------------------------------------------------------
		//---------------------------------------------------------------------------------------

		static private bool		s_tablesInitialized = false;
		static private int[]	s_lowestBitTable	= null;
		static private int[]	s_bitCountTable		= null;

		static public void InitializeTables() {
			if( s_tablesInitialized == false ) {
				s_tablesInitialized = true;
				s_lowestBitTable	= new int[256];
				s_bitCountTable		= new int[256];
				for( uint i = 0; i < 256; i ++ ) {
					s_lowestBitTable[i]	= BitUtils.LowestBitUInt32( i );
					s_bitCountTable[i]	= BitUtils.CountBitsUInt32( i );
				}
			}
		}

		//---------------------------------------------------------------------------------------
	
		static public int	LowestBitInt32_Fast( int value ) {
			Debug.Assert( s_tablesInitialized == true );
			unchecked {
				int lowestBit = s_lowestBitTable[ value & 0xff ];
				if( lowestBit != -1 ) {
					return	lowestBit;
				}
			
				lowestBit = s_lowestBitTable[ ( value >> 8 ) & 0xff ];
				if( lowestBit != -1 ) {
					return	lowestBit + 8;
				}

				lowestBit = s_lowestBitTable[ ( value >> 16 ) & 0xff ];
				if( lowestBit != -1 ) {
					return	lowestBit + 16;
				}

				lowestBit = s_lowestBitTable[ ( value >> 24 ) & 0xff ];
				if( lowestBit != -1 ) {
					return	lowestBit + 24;
				}

				return	-1;
			}
		}

		static public int	LowestBitInt20_Fast( int value ) {
			Debug.Assert( s_tablesInitialized == true );
			unchecked {
				int lowestBit = s_lowestBitTable[ value & 0xff ];
				if( lowestBit != -1 ) {
					return	lowestBit;
				}
			
				lowestBit = s_lowestBitTable[ ( value >> 8 ) & 0xff ];
				if( lowestBit != -1 ) {
					return	lowestBit + 8;
				}

				lowestBit = s_lowestBitTable[ ( value >> 16 ) & 0xff ];
				if( lowestBit != -1 ) {
					return	lowestBit + 16;
				}

				return	-1;
			}
		}

		static public int	LowestBitUInt20_Fast( uint value ) {
			Debug.Assert( s_tablesInitialized == true );
			unchecked {
				int lowestBit = s_lowestBitTable[ value & 0xff ];
				if( lowestBit != -1 ) {
					return	lowestBit;
				}
			
				lowestBit = s_lowestBitTable[ ( value >> 8 ) & 0xff ];
				if( lowestBit != -1 ) {
					return	lowestBit + 8;
				}

				lowestBit = s_lowestBitTable[ ( value >> 16 ) & 0xff ];
				if( lowestBit != -1 ) {
					return	lowestBit + 16;
				}

				return	21;
			}
		}

		static public int	LowestBitUInt32_Fast( uint value ) {
			Debug.Assert( s_tablesInitialized == true );
			unchecked {
				int lowestBit = s_lowestBitTable[ value & 0xff ];
				if( lowestBit != -1 ) {
					return	lowestBit;
				}
			
				lowestBit = s_lowestBitTable[ ( value >> 8 ) & 0xff ];
				if( lowestBit != -1 ) {
					return	lowestBit + 8;
				}

				lowestBit = s_lowestBitTable[ ( value >> 16 ) & 0xff ];
				if( lowestBit != -1 ) {
					return	lowestBit + 16;
				}

				lowestBit = s_lowestBitTable[ ( value >> 24 ) & 0xff ];
				if( lowestBit != -1 ) {
					return	lowestBit + 24;
				}

				return	-1;
			}
		}

		static public int	CountBitsInt32_Fast( int value ) {
			Debug.Assert( s_tablesInitialized == true );
			unchecked {
				return	s_bitCountTable[ value & 0xff ] +
					s_bitCountTable[ ( value >> 8 ) & 0xff ] +
					s_bitCountTable[ ( value >> 16 ) & 0xff ] +
					s_bitCountTable[ ( value >> 24 ) & 0xff ];
			}
		}

		static public int	CountBitsInt20_Fast( int value ) {
			Debug.Assert( s_tablesInitialized == true );
			unchecked {
				return	s_bitCountTable[ value & 0xff ] +
					s_bitCountTable[ ( value >> 8 ) & 0xff ] +
					s_bitCountTable[ ( value >> 16 ) & 0xff ];
			}
		}

		static public int	CountBitsUInt32_Fast( uint value ) {
			Debug.Assert( s_tablesInitialized == true );
			unchecked {
				return	s_bitCountTable[ value & 0xff ] +
					s_bitCountTable[ ( value >> 8 ) & 0xff ] +
					s_bitCountTable[ ( value >> 16 ) & 0xff ] +
					s_bitCountTable[ ( value >> 24 ) & 0xff ];
			}
		}

		//---------------------------------------------------------------------------------------
		//---------------------------------------------------------------------------------------

		static public void TestCases() {
			BitUtils.InitializeTables();

			Debug.WriteLine( BitUtils.GetBitStringUInt32( 0xffff ) );
			Debug.WriteLine( BitUtils.CountBitsUInt32( 0xffff ) );
			Debug.WriteLine( BitUtils.CountBitsUInt32_Fast( 0xffff ) );

			Debug.WriteLine( BitUtils.GetBitStringUInt32( 1234 ) );
			Debug.WriteLine( BitUtils.CountBitsUInt32( 1234 ) );
			Debug.WriteLine( BitUtils.CountBitsUInt32_Fast( 1234 ) );

			/*Debug.WriteLine( BitUtils.SetBitUInt32( 0, 0 ) );
			Debug.WriteLine( BitUtils.SetBitUInt32( 0, 1 ) );
			Debug.WriteLine( BitUtils.SetBitUInt32( 0, 2 ) );
			Debug.WriteLine( BitUtils.SetBitUInt32( 0, 3 ) );

			Debug.WriteLine( BitUtils.SetBitUInt64( 0, 0 ) );
			Debug.WriteLine( BitUtils.SetBitUInt64( 0, 1 ) );
			Debug.WriteLine( BitUtils.SetBitUInt64( 0, 2 ) );
			Debug.WriteLine( BitUtils.SetBitUInt64( 0, 3 ) );

			Debug.WriteLine( BitUtils.IsBitSetUInt64( BitUtils.SetBitUInt64( 0, 0 ), 0 ) );
			Debug.WriteLine( BitUtils.IsBitSetUInt64( BitUtils.SetBitUInt64( 0, 0 ), 1 ) );
			Debug.WriteLine( BitUtils.IsBitSetUInt64( BitUtils.SetBitUInt64( 0, 1 ), 1 ) );
			Debug.WriteLine( BitUtils.IsBitSetUInt64( BitUtils.SetBitUInt64( 0, 1 ), 2 ) );
			Debug.WriteLine( BitUtils.IsBitSetUInt64( BitUtils.SetBitUInt64( 0, 2 ), 2 ) );
			Debug.WriteLine( BitUtils.IsBitSetUInt64( BitUtils.SetBitUInt64( 0, 2 ), 3 ) );
			Debug.WriteLine( BitUtils.IsBitSetUInt64( BitUtils.SetBitUInt64( 0, 3 ), 3 ) );
			Debug.WriteLine( BitUtils.IsBitSetUInt64( BitUtils.SetBitUInt64( 0, 3 ), 4 ) );

		
			Debug.WriteLine( BitUtils.SetBitsUInt64( BitUtils.GetBitIndicesInt64( 1234567890ul ) ) );
			Debug.WriteLine( BitUtils.GetBitStringUInt64( 1234567890ul ) );
			Debug.WriteLine( BitUtils.SetBitsUInt64( BitUtils.GetBitIndicesInt64( 9999999999ul ) ) );
			Debug.WriteLine( BitUtils.GetBitStringUInt64( 9999999999ul ) );	*/
		}

		//---------------------------------------------------------------------------------------
		//---------------------------------------------------------------------------------------

	}
}
