/*
 * BSD Licence:
 * Copyright (c) 2001, Ben Houston [ ben@exocortex.org ]
 * Exocortex Technologies [ www.exocortex.org ]
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright 
 * notice, this list of conditions and the following disclaimer in the 
 * documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the <ORGANIZATION> nor the names of its contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

using System;
using System.Diagnostics;
using System.Text;

using Exocortex.LowLevel;

namespace Exocortex.Mathematics
{
	/// <summary>
	/// Summary description for NumberConverter.
	/// </summary>
	public class NumberConverter
	{
		private NumberConverter() {}

		//===============================================================================

		static public char		ToBaseDigit( int digitIndex, int digitBase ) {
			// ensure the base is valid
			if( ! ( digitBase >= 2 ) ) {
				throw new ArgumentOutOfRangeException( "digitBase", digitBase,
					"digitBase must be greater or equal to 2." );
			}
			// ensure the index of the digit is valid
			if( ! ( 0 <= digitIndex && digitIndex < digitBase ) ) {
				throw new ArgumentOutOfRangeException( "digitIndex", digitIndex,
					"digitIndex must be greater or equal to 0 and less than digitBase." );
			}

			if( digitIndex < 10 ) {
				return (char)( '0' + digitIndex );
			}
			else {
				return (char)( 'A' + digitIndex - 10 );
			}
		}

		static public char		ToBaseDigit( ulong digitIndex, int digitBase ) {
			// ensure the base is valid
			if( ! ( digitBase >= 2 ) ) {
				throw new ArgumentOutOfRangeException( "digitBase", digitBase,
					"digitBase must be greater or equal to 2." );
			}
			// ensure the index of the digit is valid
			if( ! ( 0 <= digitIndex && digitIndex < (ulong)digitBase ) ) {
				throw new ArgumentOutOfRangeException( "digitIndex", digitIndex,
					"digitIndex must be greater or equal to 0 and less than digitBase." );
			}

			if( digitIndex < 10 ) {
				return (char)( '0' + digitIndex );
			}
			else {
				return (char)( 'A' + digitIndex - 10 );
			}
		}

		//-------------------------------------------------------------------------------
		
		static public char	ToBinaryDigit( int value ) {
			return	NumberConverter.ToBaseDigit( value, 2 );
		}

		static public char	ToOctalDigit( int value ) {
			return	NumberConverter.ToBaseDigit( value, 8 );
		}

		static public char	ToDecimalDigit( int value ) {
			return	NumberConverter.ToBaseDigit( value, 10 );
		}

		static public char	ToHexadecimalDigit( int value ) {
			return	NumberConverter.ToBaseDigit( value, 16 );
		}
		
		//===============================================================================
		
		static public string ToBase( double value, int digitBase ) {
			// ensure the base is valid
			if( ! ( digitBase >= 2 ) ) {
				throw new ArgumentOutOfRangeException( "digitBase", digitBase,
					"digitBase must be greater or equal to 2." );
			}

			StringBuilder digitString = new StringBuilder();
						
			if( value < 0 ) {
				digitString.Append( '-' );
			}

			if( value == 0 ) {
				return	"0";
			}

			double	dRemainder = Math.Abs( value );
			int		digitPlace = (int) Math.Floor( Math.Log( dRemainder, digitBase ) );
			
			while( digitPlace >= 0 || dRemainder != 0 ) {
				if( digitPlace == -1 ) {
					digitString.Append( '.' );
				}
				
				double	digitMultiplier = Math.Pow( digitBase, digitPlace );
				int		digitIndex = (int) Math.Floor( dRemainder / digitMultiplier );
				dRemainder -= ( digitIndex * digitMultiplier );

				digitString.Append( NumberConverter.ToBaseDigit( digitIndex, digitBase ) );

				Debug.Assert( dRemainder < digitIndex );

				digitPlace --;
			}

			return digitString.ToString();
		}

		static public string ToBase( ulong value, int digitBase ) {
			return	ToBase( value, digitBase, 0 );
		}
		
		static public string ToBase( ulong value, int digitBase, int zeroPaddedLength ) {
			// ensure the base is valid
			if( ! ( digitBase >= 2 ) ) {
				throw new ArgumentOutOfRangeException( "digitBase", digitBase,
					"digitBase must be greater or equal to 2." );
			}

			StringBuilder digitString = new StringBuilder();
						
			/*if( d < 0 ) {
				digitString.Append( '-' );
			}*/

			/*if( ul == 0 ) {
				return "0";
			}*/

			ulong	dRemainder	= value;
			int		digitPlace;
			//Debug2.WriteVar( "ul", ul );
			if( value != 0 ) {
				// TODO: digitPlace is inaccurate when there is more than 15 bits in 'value'
				digitPlace = (int) Math.Floor( Math.Log( dRemainder, digitBase ) );
			}
			else {
				digitPlace = 0;
			}
			//Debug2.WriteVar( "digitPlace", digitPlace );
			digitPlace = Math.Max( zeroPaddedLength - 1, digitPlace );
			//Debug2.WriteVar( "digitPlace", digitPlace );

			//Debug2.WriteVar( "dRemainder", dRemainder );
			
			while( dRemainder != 0 || digitPlace >= 0 ) {
				/* should never happen -- but Debug.Assert is here just in case */
				Debug.Assert( digitPlace != -1 );
				/*if( digitPlace == -1 ) {
					digitString.Append( '.' );
				}*/
				
				ulong		digitMultiplier = (ulong) Math.Pow( digitBase, digitPlace );
				//Debug2.WriteVar( "digitMultiplier", digitMultiplier );
				ulong		digitIndex = (ulong) Math.Floor( dRemainder / digitMultiplier );
				//Debug2.WriteVar( "digitIndex", digitIndex );
				dRemainder -= ( digitIndex * digitMultiplier );
				//Debug2.WriteVar( "dRemainder", dRemainder );

				digitString.Append( NumberConverter.ToBaseDigit( digitIndex, digitBase ) );

				//Debug.Assert( ul < digitIndex );

				digitPlace --;
			}

			return digitString.ToString();
		}

		static public string ToBase( long value, int digitBase ) {
			return	ToBase( value, digitBase, 0 );
		}

		static public string ToBase( long value, int digitBase, int zeroPaddedLength ) {
			return	( ( value < 0 ) ? "-" : "" ) + 
					ToBase( Math.Abs( value ), digitBase, zeroPaddedLength );
		}

		//-------------------------------------------------------------------------------

		// unsigned (ulong)

		static public string	ToBinary( ulong value ) {
			return	NumberConverter.ToBase( value, 2 );
		}

		static public string	ToOctal( ulong value ) {
			return	NumberConverter.ToBase( value, 8 );
		}

		static public string	ToDecimal( ulong value ) {
			return	NumberConverter.ToBase( value, 10 );
		}

		static public string	ToHexadecimal( ulong value ) {
			return	NumberConverter.ToBase( value, 16 );
		}

		static public string	ToBinary( ulong value, int zeroPaddedLength ) {
			return	NumberConverter.ToBase( value, 2 );
		}

		static public string	ToOctal( ulong value, int zeroPaddedLength ) {
			return	NumberConverter.ToBase( value, 8, zeroPaddedLength );
		}

		static public string	ToDecimal( ulong value, int zeroPaddedLength ) {
			return	NumberConverter.ToBase( value, 10, zeroPaddedLength );
		}

		static public string	ToHexadecimal( ulong value, int zeroPaddedLength ) {
			return	NumberConverter.ToBase( value, 16, zeroPaddedLength );
		}

		//-------------------------------------------------------------------------------

		// signed (long)

		static public string	ToBinary( long value ) {
			return	NumberConverter.ToBase( value, 2 );
		}

		static public string	ToOctal( long value ) {
			return	NumberConverter.ToBase( value, 8 );
		}

		static public string	ToDecimal( long value ) {
			return	NumberConverter.ToBase( value, 10 );
		}

		static public string	ToHexadecimal( long value ) {
			return	NumberConverter.ToBase( value, 16 );
		}

		static public string	ToBinary( long value, int zeroPaddedLength ) {
			return	NumberConverter.ToBase( value, 2 );
		}

		static public string	ToOctal( long value, int zeroPaddedLength ) {
			return	NumberConverter.ToBase( value, 8, zeroPaddedLength );
		}

		static public string	ToDecimal( long value, int zeroPaddedLength ) {
			return	NumberConverter.ToBase( value, 10, zeroPaddedLength );
		}

		static public string	ToHexadecimal( long value, int zeroPaddedLength ) {
			return	NumberConverter.ToBase( value, 16, zeroPaddedLength );
		}
		
		//-------------------------------------------------------------------------------

	}
}
