/*
 * BSD Licence:
 * Copyright (c) 2001, Ben Houston [ ben@exocortex.org ]
 * Exocortex Technologies [ www.exocortex.org ]
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright 
 * notice, this list of conditions and the following disclaimer in the 
 * documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the <ORGANIZATION> nor the names of its contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

using System;
using System.Diagnostics;
using System.Collections;
using System.Globalization;

using Exocortex.Text;

namespace Exocortex {

	public class String2 {

		//=========================================================================================================
		//=========================================================================================================
		
		/// <summary>
		/// d	- The day of the month. Single-digit days will not have a leading zero. 
		/// dd	- The day of the month. Single-digit days will have a leading zero. 
		/// ddd	- The abbreviated name of the day of the week, as defined in AbbreviatedDayNames. 
		/// dddd	- The full name of the day of the week, as defined in DayNames. 
		/// M	- The numeric month. Single-digit months will not have a leading zero. 
		/// MM	- The numeric month. Single-digit months will have a leading zero. 
		/// MMM	- The abbreviated name of the month, as defined in AbbreviatedMonthNames. 
		/// MMMM	- The full name of the month, as defined in MonthNames. 
		/// y	- The year without the century. If the year without the century is less than 10, the year is displayed with no leading zero. 
		/// yy	- The year without the century. If the year without the century is less than 10, the year is displayed with a leading zero. 
		/// yyyy	- The year in four digits, including the century. 
		/// gg	- The period or era. This pattern is ignored if the date to be formatted does not have an associated period or era string. 
		/// h	- The hour in a 12-hour clock. Single-digit hours will not have a leading zero. 
		/// hh	- The hour in a 12-hour clock. Single-digit hours will have a leading zero. 
		/// H	- The hour in a 24-hour clock. Single-digit hours will not have a leading zero. 
		/// HH	- The hour in a 24-hour clock. Single-digit hours will have a leading zero. 
		/// m	- The minute. Single-digit minutes will not have a leading zero. 
		/// mm	- The minute. Single-digit minutes will have a leading zero. 
		/// s	- The second. Single-digit seconds will not have a leading zero. 
		/// ss	- The second. Single-digit seconds will have a leading zero. 
		/// f	- The fraction of a second in single-digit precision. The remaining digits are truncated. 
		/// ff	- The fraction of a second in double-digit precision. The remaining digits are truncated. 
		/// fff	- The fraction of a second in three-digit precision. The remaining digits are truncated. 
		/// ffff	- The fraction of a second in four-digit precision. The remaining digits are truncated. 
		/// fffff	- The fraction of a second in five-digit precision. The remaining digits are truncated. 
		/// ffffff	- The fraction of a second in six-digit precision. The remaining digits are truncated. 
		/// fffffff	- The fraction of a second in seven-digit precision. The remaining digits are truncated. 
		/// t	- The first character in the AM/PM designator defined in AMDesignator or PMDesignator. 
		/// tt	- The AM/PM designator defined in AMDesignator or PMDesignator. 
		/// z	- The time zone offset ("+" or "-" followed by the hour only). Single-digit hours will not have a leading zero. For example, Pacific Standard Time is "-8". 
		/// zz	- The time zone offset ("+" or "-" followed by the hour only). Single-digit hours will have a leading zero. For example, Pacific Standard Time is "-08". 
		/// zzz	- The full time zone offset ("+" or "-" followed by the hour and minutes). Single-digit hours and minutes will have leading zeros. For example, Pacific Standard Time is "-08:00". 
		/// :	- The default time separator defined in TimeSeparator. 
		/// /	- The default date separator defined in DateSeparator. 
		/// %c	-  Where c is a format pattern if used alone. The "%" character can be omitted if the format pattern is combined with literal characters or other format patterns. 
		/// \c	-  Where c is any character. Displays the character literally. To display the backslash character, use "\\". 
		/// </summary>
		/// 

		static public string	ToString( DateTime dateTime, string format ) {
			return	dateTime.ToString( format, DateTimeFormatInfo.InvariantInfo );
		}

		//---------------------------------------------------------------------------------------------------------

		
		static public string	GetSimpleDate() {
			return	ToString( DateTime.Now, "MMM-dd-yyyy" );
		}
		static public string	GetSimpleTime() {
			return	ToString( DateTime.Now, "HH-mm-ss" );
		}
		static public string	GetSimpleDateTime() {
			return	ToString( DateTime.Now, "MMM-dd-yyyy-HH-mm-ss");
		}
		
		static public string	ToIsoDateTime( DateTime dateTime ) {
			return	ToString( dateTime, "yyyy-MM-dd HH:mm:ss");
		}

		static public DateTime	FromIsoDateTime( string dateTime ) {
			return	DateTime.ParseExact( dateTime, "yyyy-MM-dd HH:mm:ss", DateTimeFormatInfo.InvariantInfo );
		}
		
		//===============================================================================

		const long	cKiloByte = 1024L;
		const long	cMegaByte = 1024L*1024;
		const long	cGigaByte = 1024L*1024*1024;
		const long	cTeraByte = 1024L*1024*1024*1024;

		static public string	ToMemoryUnits( long bytes ) {
			if( bytes < cKiloByte ) {
				return	bytes + " bytes";
			}
			if( bytes < cMegaByte ) {
				return	Math2.Round( ((double)bytes) / cKiloByte, 3 ) + " KB";
			}
			if( bytes < cGigaByte ) {
				return	Math2.Round( ((double)bytes) / cMegaByte, 3 ) + " MB";
			}
			if( bytes < cTeraByte ) {
				return	Math2.Round( ((double)bytes) / cGigaByte, 3 ) + " GB";
			}

			return	Math2.Round( (double)bytes / cTeraByte, 3 ) + " TB";
		}

		static public string	ToTimeUnits( TimeSpan span ) {
			
			if( span.TotalMilliseconds < 1000 ) {
				return	Math2.Round( span.TotalMilliseconds, 3 ) + " msecs";
			}
			if( span.TotalSeconds < 60 ) {
				return	Math2.Round( span.TotalSeconds, 3 ) + " secs";
			}
			if( span.TotalMinutes < 60 ) {
				return	Math2.Round( span.TotalMinutes, 3 ) + " mins";
			}
			if( span.TotalHours < 24 ) {
				return	Math2.Round( span.TotalHours, 3 ) + " hrs";
			}
			return	Math2.Round( span.TotalDays, 3 ) + " days";
		}

		//===============================================================================

		static public string[] ToTokens( string s ) {
			Debug.Assert( s != null );
			ArrayList arraylist = new ArrayList();
			StringTokenizer tokenizer = new StringTokenizer();
			tokenizer.SetBuffer( s );
			while( tokenizer.LookAhead != null ) {
				arraylist.Add( tokenizer.GetToken() );
			}
			return	(string[]) arraylist.ToArray( typeof( string ) );
		}

		//===============================================================================

	}

}
